##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
import numpy as np
from math import sqrt
from idaes.apps.matopt.materials import Atom, Canvas, Design, LinearTiling, PlanarTiling, CubicTiling
from idaes.apps.matopt.materials.geometry import Shape, Cuboctahedron, Parallelepiped, Rhombohedron, RectPrism, Cube, \
    Cylinder, CylindricalSector
from idaes.apps.matopt.materials.lattices import FCCLattice, CubicLattice, PerovskiteLattice, DiamondLattice, \
    WurtziteLattice
from idaes.apps.matopt.materials.transform_func import ShiftFunc, ScaleFunc, RotateFunc, ReflectFunc
from idaes.apps.matopt.opt import Coef, LinearExpr, MatOptModel, SumNeighborSites, SumNeighborBonds, SumSites, SumBonds, \
    SumSiteTypes, SumBondTypes, SumSitesAndTypes, SumBondsAndTypes, SumConfs, SumSitesAndConfs, LessThan, EqualTo, \
    GreaterThan, FixedTo, Disallow, PiecewiseLinear, Implies, NegImplies, ImpliesSiteCombination, ImpliesNeighbors
from idaes.apps.matopt.util.util import areEqual
import pytest
from test_matopt_objects_construction import *


@pytest.mark.unit
def test_functionality_FCCLattice():
    lattice = test_construct_FCCLattice()
    assert lattice.areNeighbors(np.zeros(3, dtype=float), np.array([0.0, -0.5, 0.5]))
    nbs = lattice.getNeighbors(np.zeros(3, dtype=float))
    for p in nbs:
        assert lattice.areNeighbors(np.zeros(3, dtype=float), p)
    assert areEqual(lattice.IAD, sqrt(2) / 2, 1e-4)
    assert areEqual(lattice.FCC111LayerSpacing, 1 / sqrt(3), 1e-4)
    assert areEqual(lattice.FCC100LayerSpacing, 1 / 2, 1e-4)
    assert areEqual(lattice.FCC110LayerSpacing, sqrt(2) / 2, 1e-4)


@pytest.mark.unit
def test_functionality_CubicLattice():
    lattice = test_construct_CubicLattice()
    assert lattice.areNeighbors(np.zeros(3, dtype=float), np.array([1.0, 0.0, 0.0]))
    nbs = lattice.getNeighbors(np.zeros(3, dtype=float))
    for p in nbs:
        assert lattice.areNeighbors(np.zeros(3, dtype=float), p)
    assert areEqual(lattice.IAD, 1.0, 1e-4)


@pytest.mark.unit
def test_functionality_DiamondLattice():
    lattice = test_construct_DiamondLattice()
    assert lattice.areNeighbors(np.zeros(3, dtype=float), np.array([0.25, 0.25, 0.25]))
    nbs = lattice.getNeighbors(np.zeros(3, dtype=float))
    for p in nbs:
        assert lattice.areNeighbors(np.zeros(3, dtype=float), p)
    assert lattice.isASite(np.zeros(3, dtype=float))
    assert lattice.isBSite(np.array([0.25, 0.25, 0.25]))
    assert areEqual(lattice.IAD, sqrt(3) / 4, 1e-4)
    assert areEqual(lattice.getLayerSpacing('100'), 1 / 4, 1e-4)
    assert areEqual(lattice.getLayerSpacing('110'), sqrt(2) / 4, 1e-4)
    assert areEqual(lattice.getLayerSpacing('111'), sqrt(3) / 3, 1e-4)
    assert areEqual(lattice.getLayerSpacing('112'), sqrt(6) / 12, 1e-4)
    assert areEqual(lattice.getShellSpacing('100'), sqrt(8) / 4, 1e-4)
    assert areEqual(lattice.getShellSpacing('112'), sqrt(2) / 4, 1e-4)
    assert areEqual(lattice.getUniqueLayerCount('100'), 4, 1e-4)
    assert areEqual(lattice.getUniqueLayerCount('110'), 2, 1e-4)
    assert areEqual(lattice.getUniqueLayerCount('111'), 3, 1e-4)
    assert areEqual(lattice.getUniqueLayerCount('112'), 6, 1e-4)


@pytest.mark.unit
def test_functionality_WurtziteLattice():
    lattice = test_construct_WurtziteLattice()
    assert lattice.areNeighbors(np.zeros(3, dtype=float), np.array([0.0, 0.0, -3 / sqrt(24)]))
    nbs = lattice.getNeighbors(np.zeros(3, dtype=float))
    for p in nbs:
        assert lattice.areNeighbors(np.zeros(3, dtype=float), p)
    assert lattice.isASite(np.zeros(3, dtype=float))
    assert lattice.isBSite(np.array([0.0, 0.0, -3 / sqrt(24)]))
    assert areEqual(lattice.IAD, sqrt(3 / 8), 1e-4)
    assert areEqual(lattice.getLayerSpacing('0001'), sqrt(2 / 3), 1e-4)
    assert areEqual(lattice.getLayerSpacing('1100'), sqrt(3) / 2, 1e-4)
    assert areEqual(lattice.getLayerSpacing('1120'), 1, 1e-4)
    assert areEqual(lattice.getShellSpacing('0001'), sqrt(3) / 2, 1e-4)
    assert areEqual(lattice.getShellSpacing('1100'), 1, 1e-4)
    assert areEqual(lattice.getShellSpacing('1120'), 1, 1e-4)
    assert areEqual(lattice.getUniqueLayerCount('0001'), 2, 1e-4)
    assert areEqual(lattice.getUniqueLayerCount('1100'), 2, 1e-4)
    assert areEqual(lattice.getUniqueLayerCount('1120'), 1, 1e-4)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
# =============================================================================

from pyomo.environ import (Integers,
    Constraint, Var, Param, exp, log, NonNegativeReals, units as pyunits)
from idaes.core.util.constants import Constants as const
from idaes.core.util.exceptions import ConfigurationError
from pyomo.core.base.units_container import _PyomoUnit
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from idaes.core.util import scaling as iscale

# Some more information about this module
__author__ = "Miguel Zamarripa"


def global_costing_parameters(self, year=None, integer_n_units=False):
    if year is None:
        year = '2018'
    # Cost index $/year (method argument or 2018 default)
    ce_index_dic = {'2019': 680, '2018': 671.1, '2017': 567.5, '2016': 541.7,
                    '2015': 556.8, '2014': 576.1, '2013': 567.3, '2012': 584.6,
                    '2011': 585.7, '2010': 550.8}

    self.CE_index = Param(mutable=True, initialize=ce_index_dic[year],
                          doc='Chemical Engineering Plant Cost Index $ year')

    if integer_n_units not in [True, False]:
        raise ValueError('{} - integer_n_units must be'
                         ' True or False'.format(self.name))
    self.integer = integer_n_units


def _make_vars(self):
    # build generic costing variables (all costing models need these vars)
    self.base_cost_per_unit = Var(initialize=1e5,
                                  domain=NonNegativeReals,
                                  doc='Unit Base Cost cost in $')
    self.purchase_cost = Var(initialize=1e4,
                             domain=NonNegativeReals,
                             doc='Unit Purchase Cost in $')
    # check global costing argument to build an integer
    # or a continuous variable
    try:  # first check flowsheet(), then parent_blocks()
        integer = self.parent_block().flowsheet().costing.integer
    except AttributeError:
        integer = self.parent_block().parent_block().costing.integer

    if integer is True:
        domain = Integers
    else:
        domain = NonNegativeReals
    self.number_of_units = Var(initialize=1,
                               domain=domain,
                               bounds=(1, 100),
                               doc="Number of units to install - "
                               "economics of scale")
    # fix number of units by default
    self.number_of_units.fix(1)


def hx_costing(self, hx_type='U-tube',
               Mat_factor='stainless steel/stainless steel',
               length_factor='12ft'):
    '''
    Heat exchanger costing method.

    Source:
        Process and Product Design Principles: Synthesis, Analysis, and
        Evaluation
        Seider, Seader, Lewin, Windagdo, 3rd Ed. John Wiley and Sons
        Chapter 22. Cost Accounting and Capital Cost Estimation
        22.2 Cost Indexes and Capital Investment

    This method computes the purchase cost (CP) for a shell and tube heat
    exchanger (Eq. 22.43), the model computes the base cost (CB for 4 types
    of heat exchangers, such as floating head, fixed head, U-tube, and
    Kettle vaporizer), construction material factor (Mat_factor),
    pressure design factor (pressure_factor), and
    tube length correction factor (length_factor),
    using Chemical Engineering base cost index of 500.

    Purchase Cost = pressure_factor * Mat_factor * length_factor * Base Cost

    Args:
        hx_type : One of: ``'floating_head'``, ``'fixed_head'``, ``'U-tube'``
        or ``'Kettle_vap'``.
        **Default** - ``'U-tube'``

        material factor : One of: ``'stain_steel'`` or ``'carb_steel'``.
            **Default** - ``'stain_steel'``

        tube length : One of: ``'8ft'``, ``'12ft'``, ``'16ft'`` or
        ``'20ft'``.
            **Default** - ``'12ft'``
    '''
    # ------------------------ Heat Exchanger cost ------------------------
    # heat exchanger cost
    # build generic costing variables
    # (base cost, purchase cost, material factor)
    _make_vars(self)

    self.L_factor = Param(mutable=True, initialize=1.12,
                          doc='HX tube length correction factor, FL')

    self.pressure_factor = Var(initialize=1,
                               domain=NonNegativeReals,
                               doc='Pressure design factor - FP')

    self.material_factor = Var(initialize=3.5,
                               domain=NonNegativeReals,
                               doc='Construction material correction'
                               'factor - Mat_factor')

    self.hx_os = Param(mutable=True, initialize=1.1,
                       doc='HX oversize factor 1.1 to 1.5',
                       units=pyunits.ft**-2)

    # select length correction factor
    c_fl = {'8ft': 1.25, '12ft': 1.12, '16ft': 1.05, '20ft': 1.00}
    self.L_factor = c_fl[length_factor]

    # --------------------------------------------------
    # base cost calculation
#       # select heat exchanger type:
    alf1 = {'floating_head': 11.9052, 'fixed_head': 11.2927, 'U-tube': 11.3852,
            'Kettle_vap': 12.2052}
    alf2 = {'floating_head': 0.8709, 'fixed_head': 0.8228, 'U-tube': 0.9186,
            'Kettle_vap': 0.8709}
    alf3 = {'floating_head': 0.09005, 'fixed_head': 0.09861, 'U-tube': 0.09790,
            'Kettle_vap': 0.09005}

    # checking units of self.parent_block().area
    area = pyunits.convert(self.parent_block().area, to_units=pyunits.ft**2)\
        / self.number_of_units

    def hx_cost_rule(self):
        return self.base_cost_per_unit == exp(alf1[hx_type]
                                              - alf2[hx_type]
                                              * log(area*self.hx_os)
                                              + alf3[hx_type]
                                              * log(area*self.hx_os)**2)
    self.base_cost_per_unit_eq = Constraint(rule=hx_cost_rule)

    @self.Expression(doc="Base cost for all units installed")
    def base_cost(self):
        return self.base_cost_per_unit * self.number_of_units

    # ------------------------------------------------------
    # Material of construction factor Eq. 22.44 in the reference
    hx_material_factor_a_dic = {'carbon steel/carbon steel':       0.00,
                                'carbon steel/brass':              1.08,
                                'carbon steel/stainless steel':    1.75,
                                'carbon steel/monel':              2.10,
                                'carbon steel/titanium':           5.20,
                                'carbon steel/Cr-Mo steel':        1.55,
                                'Cr-Mo steel/Cr-Mo steel':         1.70,
                                'stainless steel/stainless steel': 2.70,
                                'monel/monel':                     3.30,
                                'titanium/titanium':               9.60}

    hx_material_factor_b_dic = {'carbon steel/carbon steel':       0.00,
                                'carbon steel/brass':              0.05,
                                'carbon steel/stainless steel':    0.13,
                                'carbon steel/monel':              0.13,
                                'carbon steel/titanium':           0.16,
                                'carbon steel/Cr-Mo steel':        0.05,
                                'Cr-Mo steel/Cr-Mo steel':         0.07,
                                'stainless steel/stainless steel': 0.07,
                                'monel/monel':                     0.08,
                                'titanium/titanium':               0.06}

    a = hx_material_factor_a_dic[Mat_factor]
    b = hx_material_factor_b_dic[Mat_factor]

    def hx_material_fact_rule(self):
        if Mat_factor == 'carbon steel/carbon steel':
            return self.material_factor == 1
        else:
            return self.material_factor == a + (area/100/pyunits.ft**2)**b
    self.hx_material_eqn = Constraint(rule=hx_material_fact_rule)

    # ------------------------------------------------------
    # Pressure factor calculation
    # doublecheck units (higher pressure fluid should be tube side)

    pressure = pyunits.convert(
        self.parent_block().tube.properties_in[0].pressure,
        to_units=pyunits.psi)

    # units must be in psig
    def hx_P_factor(self):
        # eq valid from 600 pisg to 3000 psig
        #    return self.pressure_factor == 0.8510 + 0.1292*(pressure/600)
        #                                + 0.0198*(pressure/600)**2
        # eq valid from 100 pisg to 2000 psig
        return self.pressure_factor == (
            0.9803 + 0.0180 * (pressure/100/pyunits.psi) +
            0.0017 * (pressure/100/pyunits.psi)**2)
    self.p_factor_eq = Constraint(rule=hx_P_factor)

    # ---------------------------------------------------------
    # purchase cost equation
    def hx_CP_rule(self):
        return self.purchase_cost == (self.pressure_factor*self.material_factor
                                      * self.L_factor
                                      * (self.parent_block().flowsheet().
                                         costing.CE_index/500)*self.base_cost)
    self.cp_cost_eq = Constraint(rule=hx_CP_rule)


def pressure_changer_costing(self, Mat_factor="stain_steel",
                             # applies for all (pump, compressor, fan, blower)
                             mover_type="compressor",
                             # fan, blower, compressor
                             compressor_type="centrifugal",
                             # only for compressor
                             driver_mover_type="electrical_motor",
                             # only for compressors
                             pump_type="centrifugal",
                             # centrifugal, external_gear, reciprocating
                             pump_type_factor='1.4',
                             # 1.1 to 1.4, 2.1 and 2.2
                             # (needs to be wise-selected by user see table)
                             pump_motor_type_factor='open',
                             # centrifugal_backward, centrifugal_straight
                             # vane_axial, tube_axial
                             fan_type='centrifugal_backward',
                             # select from table depends on fan's head
                             fan_head_factor=1.45,
                             # centrifugal and rotary
                             blower_type='centrifugal'):
    '''
    Pressure changer costing method
    Source: Process and Product Design Principles: Synthesis, Analysis,
    and Evaluation Seider, Seader, Lewin, Windagdo, 3rd Ed. John Wiley and Sons
    Chapter 22. Cost Accounting and Capital Cost Estimation
    22.2 Cost Indexes and Capital Investment

    This method computes the purchase cost (CP) for a pressure changer unit,
    and can be used for costing of pumps, fan, blower, compressor, or Turbine.

    (created 2/21/2020)

    Arguments:
        mover_type : Only if config.compressor is True. Valid values 'fan',
                        'blower', 'compressor' (default).
        compressor_type : Only if mover_type='compressor'. Valid values
                        'centrifugal' (default), 'reciprocating', 'screw'
        driver_mover_type : Only if mover_type='compressor'. Valid values
                        'electric_motor' (default), 'steam_turbine',
                        'gas_turbine'
        pump_type : Only if config.compressor is True and
                        compressor_type='pump'. Valid values 'centrifugal',
                        'external_gear', 'reciprocating'
        pump_type_factor : Only if config.compressor is True and
                        compressor_type='pump'. Valid values 1.1 to 1.4,
                        2.1 and 2.2 (default = 1.4)
        pump_motor_type_factor : Only if config.compressor is True and
                        compressor_type='pump'. Valid values 'open' (default),
                        'enclosed', 'explosion_proof''
        Mat_factor : construction material; Valid values 'stain_steel'
                        (default), 'nickel_alloy' (compressor only)

    Returns:
        None
    '''
    # build generic costing variables
    # (base cost, purchase cost, material factor)
    _make_vars(self)

    self.material_factor = Param(mutable=True, initialize=3,
                                 doc='Construction material correction factor')

    # checking units
    if self.parent_block().config.thermodynamic_assumption.name == 'pump':
        w = (self.parent_block().
             work_fluid[self.parent_block().
                        flowsheet().config.time.first()])
    else:
        w = (self.parent_block().
             work_mechanical[self.parent_block().
                             flowsheet().config.time.first()])

    work_hp = pyunits.convert(w, to_units=pyunits.hp) / self.number_of_units

    # if compressor is == False, that means pressure changer is a Turbine
    if self.parent_block().config.compressor is False:
        #                           -1*work_hp because work is negative
        def CP_rule(self):
            return self.purchase_cost == self.number_of_units *\
                530*(-1/pyunits.hp * work_hp)**0.81
        self.cp_cost_eq = Constraint(rule=CP_rule)

    # if compressor is True
    # Then the pressure changer could be a Pump, Fan, Blower, or Compressor
    else:
        # if ThermodynamicAssumption == Pump (the unit is a Pump)
        if self.parent_block().config.thermodynamic_assumption.name == 'pump':
            # assign work fluid  = to w
            w = self.parent_block().work_fluid[self.parent_block().
                                               flowsheet().config.time.first()]

            # new variables only used by pump costing
            self.pump_head = Var(
                initialize=10,
                domain=NonNegativeReals,
                doc='Pump Head in feet of fluid flowing (Pressure rise/dens)',
                units=pyunits.pound_force*pyunits.foot/pyunits.pound)

            self.size_factor = Var(
                initialize=10000,
                domain=NonNegativeReals,
                doc='Pump size factor, f(Q,pump_head)')

            self.motor_base_cost_per_unit = Var(initialize=10000,
                                                domain=NonNegativeReals,
                                                doc='Motor base purchase'
                                                ' cost in $ per unit')

            self.pump_purchase_cost = Var(initialize=100000,
                                          domain=NonNegativeReals,
                                          doc='Pump purchase cost in $')

            self.motor_purchase_cost = Var(initialize=100000,
                                           domain=NonNegativeReals,
                                           doc='Motor purchase cost in $')

            self.FT = Param(mutable=True, initialize=1,
                            doc='Pump-type factor')

            # Pressure units required in lbf/ft^2
            deltaP_lb_ft2 = pyunits.convert(
                self.parent_block().deltaP[0],
                to_units=pyunits.psi*pyunits.inch**2/pyunits.foot**2)

            # dens mass units required in lb/ft^3
            dens_mass_lb_ft3 = pyunits.convert(
                self.parent_block().control_volume.properties_in[0].dens_mass,
                to_units=pyunits.pound/pyunits.foot**3)

            # volumetric flow units required in gpm
            Q_gpm = pyunits.convert(
                self.parent_block().control_volume.properties_in[0].flow_vol,
                to_units=pyunits.gallon/pyunits.minute) / self.number_of_units

            # build pump_head equation
            def p_head_rule(self):
                return self.pump_head == deltaP_lb_ft2 / dens_mass_lb_ft3
            self.p_head_eq = Constraint(rule=p_head_rule)

            # S  = Q(H)**0.5 (S = Size factor for pump)
            # Q = is the flow rate through the pump in gallons per minute
            # H = pump head in feet of flowing (pressure rise/liquid density)
            # build Size Factor equation
            def p_s_factor_rule(self):
                return self.size_factor == (
                    Q_gpm*pyunits.minute/pyunits.gallon *
                    (self.pump_head *
                     pyunits.pound/pyunits.pound_force/pyunits.foot)**0.5)
            self.s_factor_eq = Constraint(rule=p_s_factor_rule)

            # Base cost and Purchase cost for centrifugal pump
            # material factor dictionary for
            #                       centrifugal pumps and external gear pumps
            material_factor_dic = {'cast_iron':    1.00,
                                   'ductile_iron': 1.15,
                                   'cast_steel':   1.35,
                                   'bronze':       1.90,
                                   'stain_steel':  2.00,
                                   'hastelloy_c':  2.95,
                                   'monel':        3.30,
                                   'nickel':       3.50,
                                   'titanium':     9.70}
            # material factor dictionary for Reciprocating Plunger pumps
            material_factor_reciprocal_pumps = {'ductile_iron': 1.00,
                                                'Ni_Al_Bronze': 1.15,
                                                'carbon_steel': 1.50,
                                                'stain_steel': 2.20}
            # pump type factor dictionary (only used by centrifugal pumps)
            pump_type_factor_dic = {'1.1': 1.00,
                                    '1.2': 1.50,
                                    '1.3': 1.70,
                                    '1.4': 2.00,
                                    '2.1': 2.70,
                                    '2.2': 8.90}

            if pump_type == 'centrifugal':
                self.material_factor = material_factor_dic[Mat_factor]
                self.FT = pump_type_factor_dic[pump_type_factor]

            elif pump_type == 'external_gear':
                self.material_factor = material_factor_dic[Mat_factor]
                self.FT = 1

            elif pump_type == 'reciprocating':
                self.material_factor = \
                    material_factor_reciprocal_pumps[Mat_factor]
                self.FT = 1
            else:
                raise ValueError('{} - pump type not supported. '
                                 'Please see documentation for '
                                 'supported options.'.format(self.name,
                                                             pump_type))

            # pump cost correlations ---------------------------------
            def base_pump_rule(self):
                if pump_type == 'centrifugal':
                    return self.base_cost_per_unit == \
                        exp(9.7171
                            - 0.6019*log(self.size_factor)
                            + 0.0519*log(self.size_factor)**2)
                elif pump_type == 'external_gear':
                    return self.base_cost_per_unit == \
                        exp(7.6964
                            + 0.1986*log(Q_gpm)
                            + 0.0291*log(Q_gpm)**2)

                elif pump_type == 'reciprocating':
                    # brake horsepower with efficiency np typically = 90%
                    PB = (Q_gpm*self.pump_head
                          * dens_mass_lb_ft3/7.48052)/(33000*0.90)
                    return self.base_cost_per_unit == \
                        exp(7.8103
                            + 0.26986*log(PB)
                            + 0.06718*log(PB)**2)
                else:
                    raise ValueError('{} - pump type not supported. '
                                     'Please see documentation for '
                                     'supported options.'.format(self.name,
                                                                 pump_type))
            self.base_pump_cost_per_unit_eq = Constraint(rule=base_pump_rule)

            @self.Expression(doc="Base cost for all units installed")
            def base_cost(self):
                return self.base_cost_per_unit * self.number_of_units

            def CP_pump_rule(self):
                return self.pump_purchase_cost == \
                    (self.FT * self.material_factor *
                     (self.parent_block().flowsheet().
                      costing.CE_index/394)*self.base_cost)
            self.cp_pump_cost_eq = Constraint(rule=CP_pump_rule)

            # electric motor cost correlations ------------------------------
            pump_motor_type_dic = {'open': 1,
                                   'enclosed': 1.4,
                                   'explosion_proof': 1.8}
            self.motor_FT = Param(mutable=True,
                                  initialize=pump_motor_type_dic
                                  [pump_motor_type_factor],
                                  doc='Motor type factor')

            # pump fractional efficiency
            np = (-0.316 + 0.24015*log(Q_gpm*pyunits.minute/pyunits.gallon) -
                  0.01199*log(Q_gpm*pyunits.minute/pyunits.gallon)**2)
            # fractional efficiency of the electric motor
            nm = (0.80 + 0.0319*log(work_hp/pyunits.hp) -
                  0.00182*log(work_hp/pyunits.hp)**2)

            # power consumption in horsepower
            @self.Expression()
            def power_consumption_hp(self):
                return (Q_gpm*self.pump_head
                        * dens_mass_lb_ft3/7.48052)/(33000*np*nm)

            def base_motor_cost_rule(self):
                pc_hp = (self.power_consumption_hp /
                         pyunits.get_units(self.power_consumption_hp))
                return self.motor_base_cost_per_unit == exp(
                    5.8259 + 0.13141*log(pc_hp) + 0.053255*log(pc_hp)**2 +
                    0.028628*log(pc_hp)**3 - 0.0035549*log(pc_hp)**4)
            self.base_motor_cost_eq = Constraint(rule=base_motor_cost_rule)

            @self.Expression(doc="Base cost for all units installed")
            def motor_base_cost(self):
                return self.motor_base_cost_per_unit * self.number_of_units

            def CP_motor_rule(self):
                return self.motor_purchase_cost == \
                    (self.motor_FT * self.parent_block().
                     flowsheet().costing.CE_index/394 * self.motor_base_cost)
            self.cp_motor_cost_eq = Constraint(rule=CP_motor_rule)

            # Total pump cost (pump + electrical motor)
            def cp_cost_rule(self):
                return self.purchase_cost == self.motor_purchase_cost \
                    + self.pump_purchase_cost
            self.cp_cost_eq = Constraint(rule=cp_cost_rule)
        # ends pump costing code

        # compressor = True, and using isothermal assumption
        # (costing not needed)
        elif (self.parent_block().config.
              thermodynamic_assumption.name) == 'isothermal':
            raise ValueError('{} - pressure changers with isothermal '
                             'assumption are too simple to be costed. '.
                             format(self.name, mover_type))
        # if config.compressor is = True
        # if thermodynamic_assumption is not = Pump
        # (pressure changer could be a Fan, Blower, or Compressor)
        else:
            # The user has to select mover_type [compressor or Fan or Blower]
            if mover_type == "compressor":
                # Compressor Purchase Cost Correlation
                FD_param = {'electrical_motor': 1, 'steam_turbine': 1.15,
                            'gas_turbine': 1.25}
                self.FD = Param(mutable=True,
                                initialize=FD_param[driver_mover_type],
                                doc='Mover drive factor')
                # compressor material factor dictionary
                material_factor_dic = {'carbon_steel': 1, 'stain_steel': 2.5,
                                       'nickel_alloy': 5.0}

                self.material_factor = material_factor_dic[Mat_factor]

                c_alf1 = {'centrifugal': 7.58,
                          'reciprocating': 7.9661,
                          'screw': 8.1238}

                c_alf2 = {'centrifugal': 0.8,
                          'reciprocating': 0.8,
                          'screw': 0.7243}

                # Purchase cost rule
                def CB_rule(self):
                    return self.base_cost_per_unit == \
                        exp(c_alf1[compressor_type]
                            + c_alf2[compressor_type]*log(work_hp/pyunits.hp))
                self.base_cost_per_unit_eq = Constraint(rule=CB_rule)

                @self.Expression(doc="Base cost for all units installed")
                def base_cost(self):
                    return self.base_cost_per_unit * self.number_of_units

                def CP_rule(self):
                    return self.purchase_cost == \
                        (self.FD * self.material_factor
                         * (self.parent_block().flowsheet().costing.
                            CE_index/500)*self.base_cost)
                self.cp_cost_eq = Constraint(rule=CP_rule)

            # FAN Costing Model --------------------------------------------
            elif mover_type == "fan":

                # volumetric flow units required in cfm
                if(self.parent_block().config.property_package.get_metadata().
                   properties['flow_vol']['units']) == 'm^3/s':
                    # 1 m3/s = 2118.88 cfm
                    Q_cfm = self.parent_block().control_volume.\
                        properties_in[0].flow_vol*2118.88
                elif(self.parent_block().config.property_package.
                     get_metadata().
                     properties['flow_vol']['units']) == 'm^3/hr':
                    Q_cfm = self.parent_block().control_volume.\
                        properties_in[0].flow_vol*0.589  # 1 m3/hr = 0.589cfm
                elif(self.parent_block().config.property_package.
                     get_metadata().
                     properties['flow_vol']['units']) == 'ft^3/s':
                    Q_cfm = self.parent_block().control_volume.\
                        properties_in[0].flow_vol*60  # 1ft3/s*60s/min = ft3/m
                # end volumetric flow units
                else:
                    raise ValueError('{} - volumetric flowrate units '
                                     'not supported. '
                                     'Please see documentation for list of '
                                     'supported units.'.format(self.name,
                                                               mover_type))

                # fan cost correlation
                c_alf1 = {'centrifugal_backward': 11.0757,
                          'centrifugal_straight': 12.1678,
                          'vane_axial': 9.5229,
                          'tube_axial': 6.12905}

                c_alf2 = {'centrifugal_backward': 1.12906,
                          'centrifugal_straight': 1.31363,
                          'vane_axial': 0.97566,
                          'tube_axial': 0.40254}

                c_alf3 = {'centrifugal_backward': 0.08860,
                          'centrifugal_straight': 0.09974,
                          'vane_axial': 0.08532,
                          'tube_axial': 0.05787}

                self.head_factor = fan_head_factor

                material_factor_dic = {'carbon_steel': 1.0,
                                       'fiberglass':   1.8,
                                       'stain_steel':  2.5,
                                       'nickel_alloy': 5.0}

                self.material_factor = material_factor_dic[Mat_factor]

                # Base cost
                def CB_rule(self):
                    return self.base_cost_per_unit == \
                        exp(c_alf1[fan_type] - c_alf2[fan_type]*(log(Q_cfm))
                            + c_alf3[fan_type]*(log(Q_cfm)**2))
                self.base_cost_per_unit_eq = Constraint(rule=CB_rule)

                @self.Expression(doc="Base cost for all units installed")
                def base_cost(self):
                    return self.base_cost_per_unit * self.number_of_units

                def CP_rule(self):
                    return self.purchase_cost == \
                        (self.material_factor * self.head_factor
                         * (self.parent_block().flowsheet().costing.
                            CE_index/500)*self.base_cost)
                self.cp_cost_eq = Constraint(rule=CP_rule)

            # Blower Costing -------------------------------------------------
            elif mover_type == "blower":
                # Blower Cost Correlation
                c_alf1 = {'centrifugal': 6.8929,
                          'rotary': 7.59176}

                c_alf2 = {'centrifugal': 0.7900,
                          'rotary': 0.7932}

                c_alf3 = {'centrifugal': 0.0,
                          'rotary': 0.012900}

                material_factor_dic = {'carbon_steel': 1.0,
                                       'aluminum': 0.60,
                                       'fiberglass':   1.8,
                                       'stain_steel':  2.5,
                                       'nickel_alloy': 5.0}
                self.material_factor = material_factor_dic[Mat_factor]

                # Base cost
                def CB_rule(self):
                    return self.base_cost_per_unit == \
                        exp(c_alf1[blower_type]
                            + c_alf2[blower_type]*(log(work_hp))
                            - c_alf3[blower_type]*(log(work_hp)**2))
                self.base_cost_per_unit_eq = Constraint(rule=CB_rule)

                @self.Expression(doc="Base cost for all units installed")
                def base_cost(self):
                    return self.base_cost_per_unit * self.number_of_units

                def CP_rule(self):
                    return self.purchase_cost == \
                        (self.material_factor
                         * (self.parent_block().flowsheet().costing.
                            CE_index/500)*self.base_cost)
                self.cp_cost_eq = Constraint(rule=CP_rule)
            else:
                raise ValueError('{} - mover type not supported. '
                                 'Please see documentation for list of '
                                 'supported mover types.'.format(self.name,
                                                                 mover_type))


def vessel_costing(self, alignment='horizontal',
                   # material of construction of the vessel
                   Mat_factor='carbon_steel',
                   # option 1: 1000 to 920,000 lb; option2: 4200 to 1M lb
                   weight_limit='option1',
                   # for vertical vessels only
                   # option1: 3 < D < 21 ft; 12 < L < 40 ft
                   # option2: 3 < D < 24 ft; 27 < L < 170 ft
                   L_D_range='option1',
                   # True to add platforms and ladders cost
                   PL=True,
                   # True for plates cost (for distillation column)
                   plates=False,
                   # if tray cost is True, material of construction type needed
                   tray_mat_factor='carbon_steel',
                   # if tray cost is True, provide type of the trays
                   tray_type='sieve',
                   number_tray=10,
                   ref_parameter_diameter=None,
                   ref_parameter_length=None):

    # make vars (base cost, purchase cost)
    _make_vars(self)

    if ref_parameter_diameter is None and ref_parameter_length is None:
        try:
            D = self.parent_block().diameter
        except AttributeError:
            raise ConfigurationError(
                '{} Component diameter does not exist in unit model'
                .format(self.name))
        try:
            L = self.parent_block().length
        except AttributeError:
            raise ConfigurationError(
                '{} Component length does not exist in unit model'
                .format(self.name))
    elif ref_parameter_diameter is None or ref_parameter_length is None:
        raise ConfigurationError('provide both ref_parameter_diameter '
                                 'and ref_parameter_length')
    else:
        D = ref_parameter_diameter
        L = ref_parameter_length

    D = pyunits.convert(D, to_units=pyunits.foot)
    L = pyunits.convert(L, to_units=pyunits.foot)
    D_in = pyunits.convert(D, to_units=pyunits.inch)
    L_in = pyunits.convert(L, to_units=pyunits.inch)

    # new variables and parameters
    self.weight = Var(initialize=1000,
                      domain=NonNegativeReals,
                      doc='Weight of vessel in lb',
                      units=pyunits.pound)

    self.vessel_purchase_cost = Var(initialize=1e5,
                                    domain=NonNegativeReals,
                                    doc='Vessel cost in $')

    # we recommend users to calculate the pressure design based shell thickness
    # shell thickness here already considers the pressure design
    self.shell_thickness = Param(mutable=True,
                                 initialize=1.25,
                                 doc='Shell thickness in in',
                                 units=pyunits.inch)
    self.material_factor = Param(mutable=True, initialize=3,
                                 doc='Construction material correction factor')
    self.material_density = Param(mutable=True, initialize=0.284,
                                  doc='Density of the metal in lb/in^3',
                                  units=pyunits.pound/pyunits.inch**3)

    if weight_limit == 'option1':
        # Blower Cost Correlation
        c_alf1 = {'horizontal': 8.9552,
                  'vertical': 7.0132}

        c_alf2 = {'horizontal': -0.2330,
                  'vertical': 0.18255}

        c_alf3 = {'horizontal': 0.04333,
                  'vertical': 0.02297}

    elif weight_limit == 'option2':
        # option 2 only for vertical vessels
        c_alf1 = {'vertical': 7.2756}

        c_alf2 = {'vertical': 0.18255}

        c_alf3 = {'vertical': 0.02297}

    # material factor -------------------
    material_factor_dic = {'carbon_steel': 1.0,
                           'low_alloy_steel': 1.2,
                           'stain_steel_304': 1.7,
                           'stain_steel_316': 2.1,
                           'carpenter_20CB-3': 3.2,
                           'nickel_200': 5.4,
                           'monel_400': 3.6,
                           'inconel_600': 3.9,
                           'incoloy_825': 3.7,
                           'titanium': 7.7}

    self.material_factor = material_factor_dic[Mat_factor]
    # metal densities in lb/in^3
    material_dens_dic = {
        'carbon_steel': 0.284,  # 490 lb/ft3
        'low_alloy_steel': 0.271,  # 0.292 lb/in³
        'stain_steel_304': 0.270,  # 467 - 499 lb/ft3
        'stain_steel_316': 0.276,  # 467 - 499 lb/ft3
        'carpenter_20CB-3': 0.29,  # 503 lb/ft3
        'nickel_200': 0.3216,  # 556 lb/ft3
        'monel_400': 0.319,  # 522 - 552 lb/ft3
        'inconel_600': 0.3071,  # 530 lb/ft3
        'incoloy_825': 0.2903,  # 501 lb/ft3
        'titanium': 0.1628}  # 281 lb/ft3

    self.material_density.set_value(material_dens_dic[Mat_factor])

    # weight in lb --------------------------------------------------------
    # converting D in ft to inches, 0.8*D = accounts for two heads of vessel
    def weight_rule(self):
        return self.weight * self.number_of_units == (
            const.pi * (D_in + self.shell_thickness) *
            (L_in + 0.8*D_in)*self.shell_thickness*self.material_density)
    self.weight_eq = Constraint(rule=weight_rule)

    # Base Vessel cost
    def CV_rule(self):
        return self.base_cost_per_unit == \
            exp(c_alf1[alignment] +
                c_alf2[alignment]*(log(self.weight/pyunits.pound)) +
                c_alf3[alignment]*(log(self.weight/pyunits.pound)**2))
    self.cv_cost_eq = Constraint(rule=CV_rule)

    @self.Expression(doc="Base cost for all units installed")
    def base_cost(self):
        return self.base_cost_per_unit * self.number_of_units

    # True if platform and ladder costs are incuded
    if PL is True:
        self.base_cost_platf_ladders = Var(initialize=1000,
                                           domain=NonNegativeReals,
                                           doc='Base cost of'
                                           ' platforms and ladders in $')
        platforms_ladders(self, alignment=alignment, L_D_range=L_D_range,
                          D=D, L=L)

    # purchase cost of vessel and platforms and ladders
    def CP_vessel_rule(self):
        return self.vessel_purchase_cost == \
            (self.parent_block().parent_block().costing.CE_index/500) \
            * (self.material_factor*self.base_cost
               + (self.base_cost_platf_ladders
                  * self.number_of_units if PL else 0))
    self.cp_vessel_eq = Constraint(rule=CP_vessel_rule)

    # True if platform and ladder costs are incuded
    if plates is True:
        self.purchase_cost_trays = Var(initialize=1e6,
                                       domain=NonNegativeReals,
                                       doc='Purchase cost of trays in $')
        plates_cost(self,
                    tray_mat_factor=tray_mat_factor,
                    tray_type=tray_type, D=D, number_tray=number_tray)

    def CP_cost_rule(self):
        return self.purchase_cost == self.vessel_purchase_cost \
            + (self.purchase_cost_trays
               * self.number_of_units if plates else 0)
    self.cp_cost_eq = Constraint(rule=CP_cost_rule)


def platforms_ladders(self, alignment='horizontal', L_D_range='option1',
                      D=10, L=20):
    if alignment == 'horizontal':
        def CPL_rule(self):
            return self.base_cost_platf_ladders == (
                2005*(D/pyunits.foot)**0.20294)
        self.CPL_eq = Constraint(rule=CPL_rule)

    elif alignment == 'vertical':
        def CPL_rule(self):
            if L_D_range == 'option1':
                return self.base_cost_platf_ladders == (
                    361.8*(D/pyunits.foot)**0.73960 *
                    (L/pyunits.foot)**0.70684)
            elif L_D_range == 'option2':
                return self.base_cost_platf_ladders == (
                    309.9*(D/pyunits.foot)**0.63316 *
                    (L/pyunits.foot)**0.80161)
            else:
                raise ValueError('{} - L_D_range option not supported. '
                                 'Please see documentation for list of '
                                 'supported options.'.format(self.name,
                                                             L_D_range))
        self.CPL_eq = Constraint(rule=CPL_rule)

    else:
        raise ValueError('{} - vessel alignment type not supported. '
                                 'Please see documentation for list of '
                                 'supported options.'.format(self.name,
                                                             alignment))


def plates_cost(self,
                tray_mat_factor='carbon_steel',
                tray_type='sieve', D=10, number_tray=10):

    self.base_cost_trays = Var(initialize=1e4,
                               domain=NonNegativeReals,
                               doc='Base cost of trays in $')
    self.type_tray_factor = Param(mutable=True, initialize=1.0,
                                  doc='Tray type factor')
    self.number_tray_factor = Param(mutable=True,
                                    initialize=1.0,
                                    doc='Number of trays factor')
    self.number_trays = Param(mutable=True, initialize=15.0,
                              doc='Tray type factor')
    self.tray_material_factor = Var(initialize=1.0,
                                    doc='FTM material of construction'
                                    ' factor')
    # calc number tray factor
    if number_tray > 20:
        self.number_tray_factor = 1.0
    else:
        self.number_tray_factor = 2.25/(1.0414**number_tray)

    type_tray_dic = {'sieve': 1,
                     'valve': 1.18,
                     'bubble_cap': 1.87}

    self.type_tray_factor = type_tray_dic[tray_type]
    # calculate material of construction factor
    t_alf1 = {'carbon_steel': 1,
              'stain_steel_303': 1.189,
              'stain_steel_316': 1.401,
              'carpenter_20CB-3': 1.525,
              'monel': 2.306}

    t_alf2 = {'carbon_steel': 0,
              'stain_steel_303': 0.0577,
              'stain_steel_316': 0.0724,
              'carpenter_20CB-3': 0.0788,
              'monel': 0.1120}

    # recalculating tray factor value
    # Column diameter in ft, eq. valid for 2 to 16 ft
    def mt_factor_rule(self):
        return self.tray_material_factor == (
            t_alf1[tray_mat_factor] +
            t_alf2[tray_mat_factor]*D/pyunits.foot)
    self.mt_factor_eq = Constraint(rule=mt_factor_rule)

    # base cost for trays
    def bc_tray_rule(self):
        return self.base_cost_trays == 468.00*exp(0.1739*D/pyunits.foot)
    self.bc_tray_eq = Constraint(rule=bc_tray_rule)

    # purchase cost calculation for trays
    def CP_tray_rule(self):
        return self.purchase_cost_trays == \
            (self.parent_block().parent_block().costing.CE_index/500) \
            * self.number_trays \
            * self.number_tray_factor \
            * self.type_tray_factor \
            * self.tray_material_factor \
            * self.base_cost_trays
    self.cp_tray_eq = Constraint(rule=CP_tray_rule)


def fired_heater_costing(self,
                         fired_type='fuel',
                         Mat_factor='carbon_steel',
                         ref_parameter_pressure=None,
                         ref_parameter_heat_duty=None):
    # (base cost, purchase cost)
    _make_vars(self)
    self.material_factor = Param(mutable=True, initialize=3,
                                 doc='Construction material correction factor')

    self.pressure_factor = Var(initialize=1.1,
                               domain=NonNegativeReals,
                               doc='Pressure design factor')
    if ref_parameter_pressure is None:
        # pressure in psig
        P = self.parent_block().control_volume.properties_in[0].pressure

        # Need to convert to gauge pressure
        P = pyunits.convert(P, to_units=pyunits.psi) - 14.6959
    else:
        P = pyunits.convert(ref_parameter_pressure, to_units=pyunits.psi)

    if ref_parameter_heat_duty is None:
        Q = self.parent_block().heat_duty[0]
    else:
        Q = ref_parameter_heat_duty

    Q = pyunits.convert(Q, to_units=pyunits.BTU/pyunits.hr)\
        / self.number_of_units

    # material factor -------------------
    material_factor_dic = {'carbon_steel': 1.0,
                           'Cr-Mo_alloy': 1.4,
                           'stain_steel': 1.7}

    self.material_factor = material_factor_dic[Mat_factor]

    # pressure deisgn factor calculation
    def p_factor_rule(self):
        return self.pressure_factor == (
            0.986 - 0.0035*(P/(500.00*pyunits.psi)) +
            0.0175*(P/(500.00*pyunits.psi))**2)
    self.p_factor_eq = Constraint(rule=p_factor_rule)

    def CB_rule(self):
        if fired_type == 'fuel':
            return self.base_cost_per_unit == exp(
                0.32325 + 0.766*log(Q/pyunits.BTU*pyunits.hr))
        elif fired_type == 'reformer':
            return self.base_cost_per_unit ==\
                0.859*(Q/pyunits.BTU*pyunits.hr)**0.81
        elif fired_type == 'pyrolysis':
            return self.base_cost_per_unit ==\
                0.650*(Q/pyunits.BTU*pyunits.hr)**0.81
        elif fired_type == 'hot_water':
            return self.base_cost_per_unit == exp(
                9.593-0.3769*log((Q/pyunits.BTU*pyunits.hr)) +
                0.03434*log((Q/pyunits.BTU*pyunits.hr))**2)
        elif fired_type == 'salts':
            return self.base_cost_per_unit ==\
                12.32*(Q/pyunits.BTU*pyunits.hr)**0.64
        elif fired_type == 'dowtherm_a':
            return self.base_cost_per_unit ==\
                12.74*(Q/pyunits.BTU*pyunits.hr)**0.65
        elif fired_type == 'steam_boiler':
            return self.base_cost_per_unit ==\
                0.367*(Q/pyunits.BTU*pyunits.hr)**0.77
        else:
            raise ValueError('{} - fired heater type not supported. '
                                 'Please see documentation for list of '
                                 'supported FH types.'.format(self.name,
                                                              fired_type))
    self.base_cost_per_unit_eq = Constraint(rule=CB_rule)

    @self.Expression(doc="Base cost for all units installed")
    def base_cost(self):
        return self.base_cost_per_unit * self.number_of_units

    # purchase cost fired heater
    def CP_rule(self):
        return self.purchase_cost == self.material_factor \
            * (self.parent_block().parent_block().costing.CE_index/500) \
            * self.pressure_factor * self.base_cost
    self.cp_cost_eq = Constraint(rule=CP_rule)


def cstr_costing(self, alignment='vertical', Mat_factor='carbon_steel',
                 weight_limit='option1', L_D_range='option1', PL=True):
    vessel_costing(self, alignment=alignment,
                   Mat_factor=Mat_factor,
                   weight_limit=weight_limit,
                   L_D_range=L_D_range,
                   PL=PL)


def flash_costing(self, alignment='vertical', Mat_factor='carbon_steel',
                  weight_limit='option1', L_D_range='option1', PL=True):
    vessel_costing(self, alignment=alignment,
                   Mat_factor=Mat_factor,
                   weight_limit=weight_limit,
                   L_D_range=L_D_range,
                   PL=PL)


def rstoic_costing(self, alignment='vertical', Mat_factor='carbon_steel',
                    weight_limit='option1', L_D_range='option1', PL=True):
    vessel_costing(self, alignment=alignment,
                   Mat_factor=Mat_factor,
                   weight_limit=weight_limit,
                   L_D_range=L_D_range,
                   PL=PL)


def pfr_costing(self, alignment='horizontal', Mat_factor='carbon_steel',
                    weight_limit='option1', L_D_range='option1', PL=True):
    vessel_costing(self, alignment=alignment,
                   Mat_factor=Mat_factor,
                   weight_limit=weight_limit,
                   L_D_range=L_D_range,
                   PL=PL)


def initialize(self):
    '''
    Section to initialize costing blocks for all unit models
    '''
    try:
        # only heat exchanger and Fired heater calculate pressure factor
        calculate_variable_from_constraint(self.pressure_factor,
                                           self.p_factor_eq)
    except AttributeError:
        pass
    if hasattr(self, "hx_material_eqn"):
        # section to initialize heat exchanger costing block
        calculate_variable_from_constraint(self.material_factor,
                                           self.hx_material_eqn)

    elif hasattr(self, "pump_head"):
        # section to initialize pressure changer costing block
        # only for pump and electric motor costing initialization
        # all the other pressure changers only include purchase cost and/or
        # base_cost_per_unit (compressors, turbines, fans, and blowers)
        calculate_variable_from_constraint(
            self.pump_head,
            self.p_head_eq)

        calculate_variable_from_constraint(
            self.size_factor,
            self.s_factor_eq)

        calculate_variable_from_constraint(
            self.base_cost_per_unit,
            self.base_pump_cost_per_unit_eq
              )

        calculate_variable_from_constraint(
            self.pump_purchase_cost,
            self.cp_pump_cost_eq
              )

        calculate_variable_from_constraint(
            self.motor_base_cost_per_unit,
            self.base_motor_cost_eq
              )
        calculate_variable_from_constraint(
            self.motor_purchase_cost,
            self.cp_motor_cost_eq
              )

        calculate_variable_from_constraint(
            self.motor_purchase_cost,
            self.cp_motor_cost_eq
              )

        calculate_variable_from_constraint(
            self.purchase_cost,
            self.cp_cost_eq
              )

    elif hasattr(self, "vessel_purchase_cost"):
        # section to initialize vessel costing block
        calculate_variable_from_constraint(
            self.weight,
            self.weight_eq
              )

        calculate_variable_from_constraint(
            self.base_cost_per_unit,
            self.cv_cost_eq
              )

        if hasattr(self, "base_cost_platf_ladders "):
            calculate_variable_from_constraint(
                self.base_cost_platf_ladders,
                self.CPL_eq
                  )

        if hasattr(self, "base_cost_trays"):
            calculate_variable_from_constraint(
                self.tray_material_factor,
                self.mt_factor_eq
                  )
            calculate_variable_from_constraint(
                self.base_cost_trays,
                self.bc_tray_eq
                  )
            calculate_variable_from_constraint(
                self.purchase_cost_trays,
                self.cp_tray_eq
                  )
        # calculate CPL cost before vessel purchase cost eq
        calculate_variable_from_constraint(
            self.vessel_purchase_cost,
            self.cp_vessel_eq
              )

    try:
        # base_cost_units variable doesn't exist in turbine costing
        calculate_variable_from_constraint(
                self.base_cost_per_unit,
                self.base_cost_per_unit_eq)
    except AttributeError:
        pass

    # all costing methods use purchase cost
    calculate_variable_from_constraint(
            self.purchase_cost,
            self.cp_cost_eq)


def calculate_scaling_factors(self):
    try:
        # only turbine model doesn't include base cost
        # scaling base_cost_per_unit
        sf_a = iscale.get_scaling_factor(self.base_cost_per_unit,
                                         default=1e-4,
                                         warning=True)
        iscale.constraint_scaling_transform(
            self.base_cost_per_unit_eq, sf_a, overwrite=False)
    except AttributeError:
        pass
    # scaling purchase cost (both var/eqn exist in all costing methods)
    s_purchase_cost = iscale.get_scaling_factor(self.purchase_cost,
                                                default=1e-4,
                                                warning=True)

    iscale.constraint_scaling_transform(
        self.cp_cost_eq, s_purchase_cost, overwrite=False)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for turbine multistage model.

Author: John Eslick
"""
import pytest

import pyomo.environ as pyo
from pyomo.network import Arc
from idaes.core import FlowsheetBlock
from idaes.generic_models.unit_models import Heater
from idaes.power_generation.unit_models.helm import HelmTurbineMultistage
from idaes.generic_models.properties import iapws95
from idaes.core.util.model_statistics import (
    degrees_of_freedom, activated_equalities_generator)
import idaes.core.util.scaling as iscale
from idaes.core.util import get_solver

# Set up solver
solver = get_solver()


@pytest.mark.unit
def build_turbine_for_run_test():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    # roughly based on NETL baseline studies
    m.fs.turb = HelmTurbineMultistage(default={
        "property_package": m.fs.properties,
        "num_hp": 7,
        "num_ip": 14,
        "num_lp": 11,
        "hp_split_locations": [4,7],
        "ip_split_locations": [5, 14],
        "lp_split_locations": [4,7,9,11],
        "hp_disconnect": [7],
        "ip_split_num_outlets": {14:3}})

    # Add reheater
    m.fs.reheat = Heater(default={"property_package": m.fs.properties})
    m.fs.hp_to_reheat = Arc(source=m.fs.turb.hp_split[7].outlet_1,
                            destination=m.fs.reheat.inlet)
    m.fs.reheat_to_ip = Arc(source=m.fs.reheat.outlet,
                            destination=m.fs.turb.ip_stages[1].inlet)
    return m


@pytest.mark.component
def test_initialize():
    """Make a turbine model and make sure it doesn't throw exception"""
    m = build_turbine_for_run_test()
    turb = m.fs.turb

    # Set the inlet of the turbine
    p = 2.4233e7
    hin = pyo.value(iapws95.htpx(T=880*pyo.units.K, P=p*pyo.units.Pa))
    m.fs.turb.inlet_split.inlet.enth_mol[0].fix(hin)
    m.fs.turb.inlet_split.inlet.flow_mol[0].fix(26000)
    m.fs.turb.inlet_split.inlet.pressure[0].fix(p)

    # Set the inlet of the ip section, which is disconnected
    # here to insert reheater
    p = 7.802e+06
    hin = pyo.value(iapws95.htpx(T=880*pyo.units.K, P=p*pyo.units.Pa))
    m.fs.turb.ip_stages[1].inlet.enth_mol[0].value = hin
    m.fs.turb.ip_stages[1].inlet.flow_mol[0].value = 25220.0
    m.fs.turb.ip_stages[1].inlet.pressure[0].value = p

    for i, s in turb.hp_stages.items():
        s.ratioP[:] = 0.88
        s.efficiency_isentropic[:] = 0.9
    for i, s in turb.ip_stages.items():
        s.ratioP[:] = 0.85
        s.efficiency_isentropic[:] = 0.9
    for i, s in turb.lp_stages.items():
        s.ratioP[:] = 0.82
        s.efficiency_isentropic[:] = 0.9

    turb.hp_split[4].split_fraction[0,"outlet_2"].fix(0.03)
    turb.hp_split[7].split_fraction[0,"outlet_2"].fix(0.03)
    turb.ip_split[5].split_fraction[0,"outlet_2"].fix(0.04)
    turb.ip_split[14].split_fraction[0,"outlet_2"].fix(0.04)
    turb.ip_split[14].split_fraction[0,"outlet_3"].fix(0.15)
    turb.lp_split[4].split_fraction[0,"outlet_2"].fix(0.04)
    turb.lp_split[7].split_fraction[0,"outlet_2"].fix(0.04)
    turb.lp_split[9].split_fraction[0,"outlet_2"].fix(0.04)
    turb.lp_split[11].split_fraction[0,"outlet_2"].fix(0.04)

    # Congiure with reheater for a full test
    turb.ip_stages[1].inlet.fix()
    for i in turb.throttle_valve:
        turb.throttle_valve[i].Cv.fix()
        turb.throttle_valve[i].valve_opening.fix()
    turb.inlet_split.inlet.flow_mol.unfix()
    turb.inlet_mix.use_equal_pressure_constraint()

    iscale.calculate_scaling_factors(m)
    turb.initialize(outlvl=1)
    turb.ip_stages[1].inlet.unfix()


    for t in m.fs.time:
        m.fs.reheat.inlet.flow_mol[t].value = \
            pyo.value(turb.hp_split[7].outlet_1_state[t].flow_mol)
        m.fs.reheat.inlet.enth_mol[t].value = \
            pyo.value(turb.hp_split[7].outlet_1_state[t].enth_mol)
        m.fs.reheat.inlet.pressure[t].value = \
            pyo.value(turb.hp_split[7].outlet_1_state[t].pressure)
    m.fs.reheat.initialize(outlvl=4)
    def reheat_T_rule(b, t):
        return m.fs.reheat.control_volume.properties_out[t].temperature == 880
    m.fs.reheat.temperature_out_equation = pyo.Constraint(
            m.fs.reheat.flowsheet().config.time,
            rule=reheat_T_rule)

    pyo.TransformationFactory("network.expand_arcs").apply_to(m)
    m.fs.turb.outlet_stage.control_volume.properties_out[0].pressure.fix()

    assert degrees_of_freedom(m)==0
    solver.solve(m, tee=True)

    eq_cons = activated_equalities_generator(m)
    for c in eq_cons:
        assert abs(c.body() - c.lower) < 1e-4

    return m


@pytest.mark.component
def test_initialize_calc_cf():
    """Make a turbine model and make sure it doesn't throw exception"""
    m = build_turbine_for_run_test()
    turb = m.fs.turb

    # Set the inlet of the turbine
    p = 2.4233e7
    hin = pyo.value(iapws95.htpx(T=880*pyo.units.K, P=p*pyo.units.Pa))
    m.fs.turb.inlet_split.inlet.enth_mol[0].fix(hin)
    m.fs.turb.inlet_split.inlet.flow_mol[0].fix(26000)
    m.fs.turb.inlet_split.inlet.pressure[0].fix(p)

    # Set the inlet of the ip section, which is disconnected
    # here to insert reheater
    p = 7.802e+06
    hin = pyo.value(iapws95.htpx(T=880*pyo.units.K, P=p*pyo.units.Pa))
    m.fs.turb.ip_stages[1].inlet.enth_mol[0].value = hin
    #m.fs.turb.ip_stages[1].inlet.flow_mol[0].value = 25220.0
    m.fs.turb.ip_stages[1].inlet.pressure[0].value = p

    for i, s in turb.hp_stages.items():
        s.ratioP[:] = 0.88
        s.efficiency_isentropic[:] = 0.9
    for i, s in turb.ip_stages.items():
        s.ratioP[:] = 0.85
        s.efficiency_isentropic[:] = 0.9
    for i, s in turb.lp_stages.items():
        s.ratioP[:] = 0.82
        s.efficiency_isentropic[:] = 0.9

    turb.hp_split[4].split_fraction[0,"outlet_2"].fix(0.03)
    turb.hp_split[7].split_fraction[0,"outlet_2"].fix(0.03)
    turb.ip_split[5].split_fraction[0,"outlet_2"].fix(0.04)
    turb.ip_split[14].split_fraction[0,"outlet_2"].fix(0.04)
    turb.ip_split[14].split_fraction[0,"outlet_3"].fix(0.15)
    turb.lp_split[4].split_fraction[0,"outlet_2"].fix(0.04)
    turb.lp_split[7].split_fraction[0,"outlet_2"].fix(0.04)
    turb.lp_split[9].split_fraction[0,"outlet_2"].fix(0.04)
    turb.lp_split[11].split_fraction[0,"outlet_2"].fix(0.04)

    # Congiure with reheater for a full test
    turb.ip_stages[1].inlet.fix()
    turb.inlet_split.inlet.flow_mol.unfix()
    turb.inlet_mix.use_equal_pressure_constraint()
    for i in m.fs.turb.inlet_stage:
        m.fs.turb.inlet_stage[i].ratioP[0] = 0.6
        turb.throttle_valve[i].Cv.fix()
        turb.throttle_valve[i].valve_opening.fix()
    iscale.calculate_scaling_factors(m)
    turb.initialize(outlvl=1, calculate_inlet_cf=True, calculate_outlet_cf=True)
    turb.ip_stages[1].inlet.unfix()

    for t in m.fs.time:
        m.fs.reheat.inlet.flow_mol[t].value = \
            pyo.value(turb.hp_split[7].outlet_1_state[t].flow_mol)
        m.fs.reheat.inlet.enth_mol[t].value = \
            pyo.value(turb.hp_split[7].outlet_1_state[t].enth_mol)
        m.fs.reheat.inlet.pressure[t].value = \
            pyo.value(turb.hp_split[7].outlet_1_state[t].pressure)
    m.fs.reheat.initialize(outlvl=4)
    def reheat_T_rule(b, t):
        return m.fs.reheat.control_volume.properties_out[t].temperature == 880
    m.fs.reheat.temperature_out_equation = pyo.Constraint(
            m.fs.reheat.flowsheet().config.time,
            rule=reheat_T_rule)

    pyo.TransformationFactory("network.expand_arcs").apply_to(m)
    m.fs.turb.outlet_stage.control_volume.properties_out[0].pressure.fix()

    assert degrees_of_freedom(m)==0
    solver.solve(m, tee=True)

    eq_cons = activated_equalities_generator(m)
    for c in eq_cons:
        assert abs(c.body() - c.lower) < 1e-4

    assert pyo.value(m.fs.turb.inlet_split.inlet.flow_mol[0]) == pytest.approx(26000)

    return m



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""Tests that helmholtz specific models match generic models"""
import pytest

import pyomo.environ as pyo
import idaes.core
import idaes.generic_models.unit_models as cmodels
import idaes.power_generation.unit_models.helm as hmodels

from idaes.generic_models.properties import iapws95

@pytest.mark.component
def test_pump():
    m = pyo.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit1 = cmodels.Pump(default={"property_package": m.fs.properties})
    m.fs.unit2 = hmodels.HelmPump(default={"property_package": m.fs.properties})
    # set inputs

    Fin = 1e4 # mol/s
    hin = 4000 # J/mol
    Pin = 101325 # Pa
    Pout = 2*Pin # Pa
    eff = 0.7
    m.fs.unit1.inlet.flow_mol[0].fix(Fin)
    m.fs.unit2.inlet.flow_mol[0].fix(Fin)
    m.fs.unit1.inlet.enth_mol[0].fix(hin)
    m.fs.unit2.inlet.enth_mol[0].fix(hin)
    m.fs.unit1.inlet.pressure[0].fix(Pin)
    m.fs.unit2.inlet.pressure[0].fix(Pin)
    m.fs.unit1.outlet.pressure[0].fix(Pout)
    m.fs.unit2.outlet.pressure[0].fix(Pout)
    m.fs.unit1.efficiency_pump.fix(eff)
    m.fs.unit2.efficiency_pump.fix(eff)
    m.fs.unit1.initialize()
    m.fs.unit2.initialize()

    assert pyo.value(m.fs.unit1.control_volume.work[0]) == pytest.approx(
        pyo.value(m.fs.unit2.control_volume.work[0]), rel=1e-7)
    assert pyo.value(m.fs.unit1.control_volume.properties_out[0].temperature) == \
        pytest.approx(
            pyo.value(m.fs.unit2.control_volume.properties_out[0].temperature),
            rel=1e-7
        )


@pytest.mark.component
def test_turbine():
    m = pyo.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit1 = cmodels.Turbine(default={"property_package": m.fs.properties})
    m.fs.unit2 = hmodels.HelmIsentropicTurbine(
        default={"property_package": m.fs.properties})

    # set inputs
    Fin = 1000 # mol/s
    Tin = 500 # K
    Pin = 1e6 # Pa
    Pout = 7e5 # Pa
    hin = pyo.value(iapws95.htpx(Tin*pyo.units.K, Pin*pyo.units.Pa)) # J/mol

    m.fs.unit1.inlet.flow_mol[0].fix(Fin)
    m.fs.unit2.inlet.flow_mol[0].fix(Fin)
    m.fs.unit1.inlet.enth_mol[0].fix(hin)
    m.fs.unit2.inlet.enth_mol[0].fix(hin)
    m.fs.unit1.inlet.pressure[0].fix(Pin)
    m.fs.unit2.inlet.pressure[0].fix(Pin)
    m.fs.unit1.outlet.pressure[0].fix(Pout)
    m.fs.unit2.outlet.pressure[0].fix(Pout)
    m.fs.unit1.efficiency_isentropic.fix(0.9)
    m.fs.unit2.efficiency_isentropic.fix(0.9)
    m.fs.unit1.initialize()
    m.fs.unit2.initialize()

    assert pyo.value(m.fs.unit1.control_volume.properties_out[0].temperature) == \
        pytest.approx(
            pyo.value(m.fs.unit2.control_volume.properties_out[0].temperature),
            rel=1e-7
        )
    assert pyo.value(m.fs.unit1.control_volume.work[0]) == pytest.approx(
        pyo.value(m.fs.unit2.control_volume.work[0]), rel=1e-7)


@pytest.mark.component
def test_compressor():
    m = pyo.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit1 = cmodels.Compressor(default={"property_package": m.fs.properties})
    m.fs.unit2 = hmodels.HelmIsentropicCompressor(default={"property_package": m.fs.properties})

    # set inputs
    Fin = 1000 # mol/s
    Tin = 500 # K
    Pin = 2222782.4 # Pa
    Pout = 2.7*Pin # Pa
    hin = pyo.value(iapws95.htpx(Tin*pyo.units.K, Pin*pyo.units.Pa)) # J/mol
    eff = 0.324

    m.fs.unit1.inlet.flow_mol[0].fix(Fin)
    m.fs.unit2.inlet.flow_mol[0].fix(Fin)
    m.fs.unit1.inlet.enth_mol[0].fix(hin)
    m.fs.unit2.inlet.enth_mol[0].fix(hin)
    m.fs.unit1.inlet.pressure[0].fix(Pin)
    m.fs.unit2.inlet.pressure[0].fix(Pin)
    m.fs.unit1.outlet.pressure[0].fix(Pout)
    m.fs.unit2.outlet.pressure[0].fix(Pout)
    m.fs.unit1.efficiency_isentropic.fix(eff)
    m.fs.unit2.efficiency_isentropic.fix(eff)
    m.fs.unit1.initialize()
    m.fs.unit2.initialize()

    assert pyo.value(m.fs.unit1.control_volume.properties_out[0].temperature) == \
        pytest.approx(
            pyo.value(m.fs.unit2.control_volume.properties_out[0].temperature),
            rel=1e-7
        )
    assert pyo.value(m.fs.unit1.control_volume.work[0]) == pytest.approx(
        pyo.value(m.fs.unit2.control_volume.work[0]), rel=1e-7)

@pytest.mark.component
def test_compressor_pump_compare():
    m = pyo.ConcreteModel()
    m.fs = idaes.core.FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit1 = hmodels.HelmPump(default={"property_package": m.fs.properties})
    m.fs.unit2 = hmodels.HelmIsentropicCompressor(
        default={"property_package": m.fs.properties})

    Fin = 1e4 # mol/s
    hin = 4000 # J/mol
    Pin = 101325 # Pa
    Pout = 2*Pin # Pa
    eff = 0.7
    m.fs.unit1.inlet.flow_mol[0].fix(Fin)
    m.fs.unit1.inlet.enth_mol[0].fix(hin)
    m.fs.unit1.inlet.pressure[0].fix(Pin)
    m.fs.unit1.outlet.pressure[0].fix(Pout)
    m.fs.unit1.efficiency_pump.fix(eff)

    m.fs.unit2.inlet.flow_mol[0].fix(Fin)
    m.fs.unit2.inlet.enth_mol[0].fix(hin)
    m.fs.unit2.inlet.pressure[0].fix(Pin)
    m.fs.unit2.outlet.pressure[0].fix(Pout)
    m.fs.unit2.efficiency_isentropic.fix(eff)

    m.fs.unit1.initialize()
    m.fs.unit2.initialize()

    # The pump calculations are a bit more approximate assuming incompressible
    # fluid entropy is independent of pressure, so the results here should be
    # close, but not exactly the same.
    assert pyo.value(m.fs.unit1.control_volume.properties_out[0].temperature) == \
        pytest.approx(
            pyo.value(m.fs.unit2.control_volume.properties_out[0].temperature),
            rel=1e-3
        )
    assert pyo.value(m.fs.unit1.control_volume.work[0]) == pytest.approx(
        pyo.value(m.fs.unit2.control_volume.work[0]), rel=1e-3)



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Unit operation model for a waterpipe with pressure drop:

Created on Aug 27, 2020 by Boiler Team (J. Ma, M. Zamarripa)
"""
import pytest
# Import Pyomo libraries
import pyomo.environ as pyo

# Import IDAES core
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.power_generation.unit_models.waterpipe import WaterPipe
# Import Unit Model Modules
from idaes.generic_models.properties import iapws95

from idaes.core.util.testing import initialization_tester
from idaes.core.util import get_solver

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------


@pytest.fixture(scope="module")
def build_unit():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = WaterPipe(default={
                               "dynamic": False,
                               "property_package": m.fs.properties,
                               "has_holdup": True,
                               "has_heat_transfer": False,
                               "has_pressure_change": True,
                               "water_phase": 'Liq',
                               "contraction_expansion_at_end": 'contraction'
                               })

    return m


@pytest.mark.unit
def test_basic_build(build_unit):
    """Make a unit model and make sure it doesn't throw exception"""
    m = build_unit
    assert degrees_of_freedom(m) == 9
    # Check unit config arguments
    assert len(m.fs.unit.config) == 11
    assert m.fs.unit.config.has_pressure_change
    assert isinstance(m.fs.unit.deltaP, pyo.Var)
    assert m.fs.unit.config.property_package is m.fs.properties


@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialize_unit(build_unit):
    m = build_unit
    m.fs.unit.diameter.fix(0.04)
    m.fs.unit.length.fix(40)
    m.fs.unit.number_of_pipes.fix(100)
    m.fs.unit.elevation_change.fix(25)
    m.fs.unit.area_ratio.fix(0.5)
    m.fs.unit.fcorrection_dp.fix(1.0)

    state_args = {'flow_mol': 10000,
                  'pressure': 1.3e7,
                  'enth_mol': 18000}

    initialization_tester(build_unit, dof=3, state_args=state_args)


@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_solve_unit(build_unit):
    m = build_unit

    m.fs.unit.inlet.enth_mol.fix()
    m.fs.unit.inlet.flow_mol.fix()
    m.fs.unit.inlet.pressure.fix()
    assert degrees_of_freedom(m) == 0

    results = solver.solve(m, tee=True)
    # Check for optimal solution
    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok

    # check material balance
    assert (pytest.approx(pyo.value(m.fs.unit.control_volume.properties_in[0].
                                    flow_mol - m.fs.unit.control_volume.
                                    properties_out[0].flow_mol),
                          abs=1e-3) == 0)

    # pressure drop
    assert (pytest.approx(-224074.4039, abs=1e-3) == pyo.value(m.fs.unit.
                                                               deltaP[0]))
    # check energy balance
    assert (pytest.approx(
        pyo.value(m.fs.unit.control_volume.properties_in[0].enth_mol
                  - m.fs.unit.control_volume.properties_out[0].enth_mol),
        abs=1e-3) == 0)


@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_pipe_expansion():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = WaterPipe(default={
                               "dynamic": False,
                               "property_package": m.fs.properties,
                               "has_holdup": True,
                               "has_heat_transfer": False,
                               "has_pressure_change": True,
                               "water_phase": 'Liq',
                               "contraction_expansion_at_end": 'expansion'
                               })
    m.fs.unit.diameter.fix(0.04)
    m.fs.unit.length.fix(40)
    m.fs.unit.number_of_pipes.fix(100)
    m.fs.unit.elevation_change.fix(25)
    m.fs.unit.area_ratio.fix(0.5)
    m.fs.unit.fcorrection_dp.fix(1.0)

    state_args = {'flow_mol': 10000,
                  'pressure': 1.3e7,
                  'enth_mol': 18000}

    initialization_tester(m, dof=3, state_args=state_args)
    m.fs.unit.inlet.enth_mol.fix()
    m.fs.unit.inlet.flow_mol.fix()
    m.fs.unit.inlet.pressure.fix()
    assert degrees_of_freedom(m) == 0

    results = solver.solve(m, tee=True)
    # Check for optimal solution
    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok

    # check material balance
    assert (pytest.approx(pyo.value(m.fs.unit.control_volume.properties_in[0].
                                    flow_mol - m.fs.unit.control_volume.
                                    properties_out[0].flow_mol),
                          abs=1e-3) == 0)

    # pressure drop
    assert (pytest.approx(-224306.548, abs=1e-3) == pyo.value(m.fs.unit.
                                                              deltaP[0]))
    # check energy balance
    assert (pytest.approx(
        pyo.value(m.fs.unit.control_volume.properties_in[0].enth_mol
                  - m.fs.unit.control_volume.properties_out[0].enth_mol),
        abs=1e-3) == 0)


@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_pipe_noexpansion():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = WaterPipe(default={
                               "dynamic": False,
                               "property_package": m.fs.properties,
                               "has_holdup": True,
                               "has_heat_transfer": False,
                               "has_pressure_change": True,
                               "water_phase": 'Liq',
                               "contraction_expansion_at_end": 'None'
                               })
    m.fs.unit.diameter.fix(0.04)
    m.fs.unit.length.fix(40)
    m.fs.unit.number_of_pipes.fix(100)
    m.fs.unit.elevation_change.fix(25)
    m.fs.unit.fcorrection_dp.fix(1.0)

    state_args = {'flow_mol': 10000,
                  'pressure': 1.3e7,
                  'enth_mol': 18000}

    initialization_tester(m, dof=3, state_args=state_args)
    m.fs.unit.inlet.enth_mol.fix()
    m.fs.unit.inlet.flow_mol.fix()
    m.fs.unit.inlet.pressure.fix()
    assert degrees_of_freedom(m) == 0

    results = solver.solve(m, tee=True)
    # Check for optimal solution
    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok

    # check material balance
    assert (pytest.approx(pyo.value(m.fs.unit.control_volume.properties_in[0].
                                    flow_mol - m.fs.unit.control_volume.
                                    properties_out[0].flow_mol),
                          abs=1e-3) == 0)

    # pressure drop
    assert (pytest.approx(-219382.990, abs=1e-3) == pyo.value(m.fs.unit.
                                                              deltaP[0]))
    # check energy balance
    assert (pytest.approx(
        pyo.value(m.fs.unit.control_volume.properties_in[0].enth_mol
                  - m.fs.unit.control_volume.properties_out[0].enth_mol),
        abs=1e-3) == 0)


@pytest.mark.skipif(not iapws95.iapws95_available(),
                    reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_pipe_vaporphase():
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.unit = WaterPipe(default={
                               "dynamic": False,
                               "property_package": m.fs.properties,
                               "has_holdup": True,
                               "has_heat_transfer": False,
                               "has_pressure_change": True,
                               "water_phase": 'Vap',
                               "contraction_expansion_at_end": 'None'
                               })
    m.fs.unit.diameter.fix(0.04)
    m.fs.unit.length.fix(40)
    m.fs.unit.number_of_pipes.fix(100)
    m.fs.unit.elevation_change.fix(25)
    m.fs.unit.fcorrection_dp.fix(1.0)

    state_args = {'flow_mol': 10000,
                  'pressure': 3.6e6,
                  'enth_mol': 53942}

    initialization_tester(m, dof=3, state_args=state_args)
    m.fs.unit.inlet.enth_mol.fix()
    m.fs.unit.inlet.flow_mol.fix()
    m.fs.unit.inlet.pressure.fix()
    assert degrees_of_freedom(m) == 0

    results = solver.solve(m, tee=True)
    # Check for optimal solution
    assert results.solver.termination_condition == \
        pyo.TerminationCondition.optimal
    assert results.solver.status == pyo.SolverStatus.ok

    assert pyo.value(m.fs.unit.control_volume.properties_in[0].vapor_frac) == 1

    # check material balance
    assert (pytest.approx(pyo.value(m.fs.unit.control_volume.properties_in[0].
                                    flow_mol - m.fs.unit.control_volume.
                                    properties_out[0].flow_mol),
                          abs=1e-3) == 0)

    # pressure drop
    assert (pytest.approx(-539105.588, abs=1e-3) == pyo.value(m.fs.unit.
                                                              deltaP[0]))
    # check energy balance
    assert (pytest.approx(
        pyo.value(m.fs.unit.control_volume.properties_in[0].enth_mol
                  - m.fs.unit.control_volume.properties_out[0].enth_mol),
        abs=1e-3) == 0)



# -*- coding: utf-8 -*-
##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
""" Functions for categorizing variable into e.g. differential and algebraic.
"""

from pyomo.environ import (
        Reference,
        )
from pyomo.dae import DerivativeVar
from pyomo.common.collections import ComponentSet, ComponentMap

import idaes.apps.caprese.nmpc_var as nmpc_var
from idaes.apps.caprese.common.config import VariableCategory

CATEGORY_TYPE_MAP = {
        VariableCategory.DIFFERENTIAL: nmpc_var.DiffVar,
        VariableCategory.ALGEBRAIC: nmpc_var.AlgVar,
        VariableCategory.DERIVATIVE: nmpc_var.DerivVar,
        VariableCategory.INPUT: nmpc_var.InputVar,
        VariableCategory.FIXED: nmpc_var.FixedVar,
        VariableCategory.MEASUREMENT: nmpc_var.MeasuredVar,
        }

def categorize_dae_variables(dae_vars, time, inputs, measurements=None):
    t0 = time.first()
    t1 = time.get_finite_elements()[1]
    deriv_vars = []
    diff_vars = []
    input_vars = []
    alg_vars = []
    fixed_vars = []

    # TODO: give user ability to specify measurements and disturbances
    measured_vars = []

    if measurements is not None:
        infer_measurements = False
        user_measurements = ComponentSet(measurements)
        updated_user_measurements = ComponentSet(measurements)
        user_measured_vars = []
    else:
        infer_measurements = True
        updated_user_measurements = ComponentSet()

    dae_map = ComponentMap([(v[t0], v) for v in dae_vars])
    t0_vardata = list(dae_map.keys())

    if inputs is None:
        inputs = []
    input_set = ComponentSet(inputs)
    updated_input_set = ComponentSet(inputs)

    for var0 in t0_vardata:
        if var0 in input_set:
            updated_input_set.remove(var0)
            time_slice = dae_map.pop(var0)
            input_vars.append(time_slice)

        if var0 in updated_user_measurements:
            updated_user_measurements.remove(var0)
            # Don't pop measured vars. They will be popped elsewhere.
            time_slice = dae_map[var0]
            user_measured_vars.append(time_slice)

        parent = var0.parent_component()
        if not isinstance(parent, DerivativeVar):
            continue
        if not time in ComponentSet(parent.get_continuousset_list()):
            continue
        index0 = var0.index()
        var1 = dae_map[var0][t1]
        index1 = var1.index()
        state = parent.get_state_var()

        if state[index1].fixed:
            # Assume state var is fixed everywhere, so derivative
            # 'isn't really' a derivative.
            # Should be safe to remove state from dae_map here
            state_slice = dae_map.pop(state[index0])
            fixed_vars.append(state_slice)
            continue
        if state[index0] in input_set:
            # If differential variable is an input, then this DerivativeVar
            # is 'not really a derivative'
            continue

        deriv_slice = dae_map.pop(var0)

        if var1.fixed:
            # Assume derivative has been fixed everywhere.
            # Add to list of fixed variables, and don't remove its state variable.
            fixed_vars.append(deriv_slice)
        elif var0.fixed:
            # In this case the derivative has been used as an initial condition. 
            # Still want to include it in the list of derivatives.
            measured_vars.append(deriv_slice)
            state_slice = dae_map.pop(state[index0])
            if state[index0].fixed:
                measured_vars.append(state_slice)
            deriv_vars.append(deriv_slice)
            diff_vars.append(state_slice)
        else:
            # Neither is fixed. This should be the most common case.
            state_slice = dae_map.pop(state[index0])
            if state[index0].fixed:
                measured_vars.append(state_slice)
            deriv_vars.append(deriv_slice)
            diff_vars.append(state_slice)

    if updated_input_set:
        raise RuntimeError('Not all inputs could be found')
    assert len(deriv_vars) == len(diff_vars)

    for var0, time_slice in dae_map.items():
        var1 = time_slice[t1]
        # If the variable is still in the list of time-indexed vars,
        # it must either be fixed (not a var) or be an algebraic var
        if var1.fixed:
            fixed_vars.append(time_slice)
        else:
            if var0.fixed:
                measured_vars.append(time_slice)
            alg_vars.append(time_slice)

    category_list_map = {
            VariableCategory.DERIVATIVE: deriv_vars,
            VariableCategory.DIFFERENTIAL: diff_vars,
            VariableCategory.ALGEBRAIC: alg_vars,
            VariableCategory.INPUT: input_vars,
            VariableCategory.FIXED: fixed_vars,
            VariableCategory.MEASUREMENT: measured_vars,
            }
    if measurements is not None:
        # If the user provided their own measurements,
        # override the inferred measurements. Assume the user
        # will modify the state of their variables appropriately.
        category_list_map[VariableCategory.MEASUREMENT] = user_measured_vars
    category_dict = {
            category: [
                Reference(ref.referent, ctype=ctype)
                for ref in category_list_map[category]
                ]
            for category, ctype in CATEGORY_TYPE_MAP.items()
            }
    return category_dict





__author__ = "John Eslick"

import pyomo.environ as pyo
from pyomo.network import Arc
from idaes.power_generation.unit_models.helm import HelmValve
from idaes.core import FlowsheetBlock, MaterialBalanceType
from idaes.generic_models.unit_models import Heater
from idaes.generic_models.properties import iapws95
from idaes.core.util import copy_port_values as _set_port, get_solver


def create_model_steady_state(f=100, p=5e5, h=5e4):
    """ Create a steady state heater model.
    """
    m = pyo.ConcreteModel(name="Dynamic Heater Test")
    m.fs = FlowsheetBlock(default={"dynamic": False})
    # Create a property parameter block
    m.fs.prop_water = iapws95.Iapws95ParameterBlock(
        default={"phase_presentation": iapws95.PhaseType.MIX})
    m.fs.heater = Heater(default={
        "has_holdup": False,
        "has_pressure_change": True,
        "material_balance_type": MaterialBalanceType.componentTotal,
        "property_package": m.fs.prop_water})

    m.fs.heater.inlet.enth_mol.fix(50000)
    m.fs.heater.inlet.pressure.fix(5e5)
    m.fs.heater.inlet.flow_mol.fix(100)
    m.fs.heater.heat_duty.fix(0)
    m.fs.heater.deltaP.fix(0)

    m.fs.heater.initialize()

    solver = get_solver(options={"max_iter": 25})
    return m, solver


def create_model_dynamic(
        p_in=5e5,
        p_out=101325,
        h=5e4,
        time_set=None,
        nfe=10,
        dae_transform='dae.collocation',
        dae_scheme='LAGRANGE-RADAU'
        ):
    """ Create a test dynamic heater model and solver.  For dynamic testing
    a valve is added to the heater outlet to set up pressure driven flow.

    Args:
        time_set (list): The begining and end point of the time domain
        nfe (int): Number of finite elements argument for the DAE transformation
    Returns:
        (tuple): (ConcreteModel, Solver)
    """
    if time_set is None:
        time_set = [0,5]

    m = pyo.ConcreteModel(name="Dynamic Heater Test")
    m.fs = FlowsheetBlock(default={"dynamic":True, "time_set":time_set})
    # Create a property parameter block
    m.fs.prop_water = iapws95.Iapws95ParameterBlock(
        default={"phase_presentation":iapws95.PhaseType.MIX})
    # Create the valve and heater models
    m.fs.pipe = HelmValve(default={
        "dynamic":False,
        "has_holdup":False,
        "material_balance_type":MaterialBalanceType.componentTotal,
        "property_package":m.fs.prop_water})
    m.fs.heater = Heater(default={
        "has_holdup":True,
        "material_balance_type":MaterialBalanceType.componentTotal,
        "property_package":m.fs.prop_water})
    m.fs.valve = HelmValve(default={
        "dynamic":False,
        "has_holdup":False,
        "material_balance_type":MaterialBalanceType.componentTotal,
        "property_package":m.fs.prop_water})

    # Connect the models
    m.fs.v1_to_t = Arc(source=m.fs.pipe.outlet, destination=m.fs.heater.inlet)
    m.fs.t_to_v2 = Arc(source=m.fs.heater.outlet, destination=m.fs.valve.inlet)

    # Add the stream constraints and do the DAE transformation
    pyo.TransformationFactory('network.expand_arcs').apply_to(m.fs)
    pyo.TransformationFactory('dae.finite_difference').apply_to(
        m.fs, nfe=nfe, wrt=m.fs.time, scheme='BACKWARD')

    # Fix the derivative variables to zero at time 0 (steady state assumption)
    m.fs.fix_initial_conditions()

    # A heater pressure reference that's directly time-indexed
    m.fs.heater_pressure = pyo.Reference(
        m.fs.heater.control_volume.properties_out[:].pressure)

    # Fix the input variables
    m.fs.pipe.inlet.enth_mol.fix(h)
    m.fs.pipe.inlet.pressure.fix(p_in)
    m.fs.valve.outlet.pressure.fix(p_out)
    m.fs.pipe.Cv.fix(0.001)
    m.fs.valve.Cv.fix(0.001)
    m.fs.pipe.valve_opening.fix(1)
    m.fs.valve.valve_opening.fix(1)
    m.fs.heater.heat_duty.fix(0)
    m.fs.heater.control_volume.volume.fix(2.0)

    # Initialize the model
    solver = get_solver(options={"max_iter": 25})
    for t in m.fs.time:
        m.fs.pipe.inlet.flow_mol = 250 # initial guess on flow
    # simple initialize
    m.fs.pipe.initialize()
    _set_port(m.fs.heater.inlet, m.fs.pipe.outlet)
    m.fs.heater.initialize()
    _set_port(m.fs.valve.inlet, m.fs.heater.outlet)
    m.fs.valve.initialize()
    solver.solve(m, tee=True)

    # Return the model and solver
    return m, solver



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for turbine inlet model.

Author: John Eslick
"""
import pytest

from pyomo.environ import (ConcreteModel, SolverFactory, TransformationFactory,
                           units as pyunits, value)

from idaes.core import FlowsheetBlock
from idaes.power_generation.unit_models.helm import HelmTurbineInletStage
from idaes.generic_models.properties import iapws95
from idaes.core.util.model_statistics import (
        degrees_of_freedom,
        activated_equalities_generator)
from idaes.core.util import get_solver

# Set up solver
solver = get_solver()


@pytest.fixture()
def build_turbine():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.turb = HelmTurbineInletStage(default={"property_package": m.fs.properties})
    return m


@pytest.fixture()
def build_turbine_dyn():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": True, "time_set":[0,4]})
    m.fs.properties = iapws95.Iapws95ParameterBlock()
    m.fs.turb = HelmTurbineInletStage(default={"property_package": m.fs.properties})
    return m


@pytest.mark.unit
def test_basic_build(build_turbine):
    """Make a turbine model and make sure it doesn't throw exception"""
    m = build_turbine


@pytest.mark.component
def test_initialize(build_turbine):
    """Initialize a turbine model"""
    m = build_turbine
    hin = value(iapws95.htpx(T=880*pyunits.K, P=2.4233e7*pyunits.Pa))
    # set inlet
    m.fs.turb.inlet.enth_mol[0].value = hin
    m.fs.turb.inlet.flow_mol[0].value = 26000/4.0
    m.fs.turb.inlet.pressure[0].value = 2.4233e7
    m.fs.turb.initialize(outlvl=1)
    eq_cons = activated_equalities_generator(m)
    for c in eq_cons:
        assert abs(c.body() - c.lower) < 1e-4
    assert degrees_of_freedom(m)==3 #inlet was't fixed and still shouldn't be


@pytest.mark.component
def test_initialize_calc_cf(build_turbine):
    """Initialize a turbine model"""
    m = build_turbine
    hin = value(iapws95.htpx(T=880*pyunits.K, P=2.4233e7*pyunits.Pa))
    # set inlet
    m.fs.turb.inlet.enth_mol[0].value = hin
    m.fs.turb.inlet.flow_mol[0].value = 26000/4.0
    m.fs.turb.inlet.pressure[0].value = 2.4233e7
    m.fs.turb.ratioP[0].value = 0.6

    m.fs.turb.initialize(outlvl=1, calculate_cf=True)
    eq_cons = activated_equalities_generator(m)
    for c in eq_cons:
        assert abs(c.body() - c.lower) < 1e-4

    m.fs.turb.inlet.enth_mol[0].value = hin
    m.fs.turb.inlet.flow_mol[0].value = 26000/4.0
    m.fs.turb.inlet.pressure[0].value = 2.4233e7
    m.fs.turb.inlet.fix()
    solver.solve(m)
    assert m.fs.turb.ratioP[0].value == pytest.approx(0.6)

    assert degrees_of_freedom(m)==0


@pytest.mark.component
def test_initialize_dyn(build_turbine_dyn):
    """Initialize a turbine model"""
    m = build_turbine_dyn
    hin = value(iapws95.htpx(T=880*pyunits.K, P=2.4233e7*pyunits.Pa))
    discretizer = TransformationFactory('dae.finite_difference')
    discretizer.apply_to(m, nfe=4, wrt=m.fs.time, scheme='BACKWARD')

    # fix inlet
    m.fs.turb.inlet.enth_mol.fix(hin)
    m.fs.turb.inlet.flow_mol.fix(26000/4.0)
    m.fs.turb.inlet.pressure.fix(2.4233e7)
    m.fs.turb.flow_coeff[:].fix()

    m.fs.turb.initialize()

    eq_cons = activated_equalities_generator(m)
    for c in eq_cons:
        assert(abs(c.body() - c.lower) < 1e-4)

    assert(degrees_of_freedom(m)==0)


@pytest.mark.component
def test_initialize_dyn2(build_turbine_dyn):
    """Initialize a turbine model"""
    m = build_turbine_dyn
    hin = iapws95.htpx(T=880*pyunits.K, P=2.4233e7*pyunits.Pa)
    discretizer = TransformationFactory('dae.finite_difference')
    discretizer.apply_to(m, nfe=4, wrt=m.fs.time, scheme='BACKWARD')

    # fix inlet
    m.fs.turb.inlet.enth_mol.fix(hin)
    m.fs.turb.inlet.flow_mol.fix(26000/4.0)
    m.fs.turb.inlet.pressure.fix(2.4233e7)
    m.fs.turb.flow_coeff[:].fix()
    m.fs.turb.ratioP[0].value = 0.6 # used for flow coeff calc

    m.fs.turb.initialize(calculate_cf=True)

    eq_cons = activated_equalities_generator(m)
    for c in eq_cons:
        assert(abs(c.body() - c.lower) < 1e-4)
    assert m.fs.turb.ratioP[0].value == pytest.approx(0.6)
    assert(degrees_of_freedom(m)==0)


@pytest.mark.unit
def test_report(build_turbine):
    m = build_turbine
    m.fs.turb.report()



##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
"""
Tests for flowsheet_model.

Author: Andrew Lee
"""
import pytest
import types

from pyomo.environ import ConcreteModel, Constraint, Set, Var
from pyomo.common.config import ConfigBlock

from idaes.core import (declare_process_block_class, PhysicalParameterBlock,
                        StateBlock, StateBlockData)
from idaes.core.phases import Phase
from idaes.core.components import Component
from idaes.core.util.exceptions import (PropertyPackageError,
                                        PropertyNotSupportedError)
from idaes.core.property_meta import PropertyClassMetadata


# -----------------------------------------------------------------------------
# Test ParameterBlock
@declare_process_block_class("ParameterBlock")
class _ParameterBlock(PhysicalParameterBlock):
    pass


@pytest.mark.unit
def test_config_block():
    # Test that PhysicalParameterBlock gets module information
    m = ConcreteModel()
    m.p = ParameterBlock()

    assert len(m.p.config) == 1
    assert isinstance(m.p.config.default_arguments, ConfigBlock)
    assert len(m.p.config.default_arguments) == 0


@pytest.mark.unit
def test_PhysicalParameterBlock():
    # Test that PhysicalParameterBlock builds correctly
    m = ConcreteModel()
    m.p = ParameterBlock()
    super(_ParameterBlock, m.p).build()


@pytest.mark.unit
def test_PhysicalParameter_NotImplementedErrors():
    # Test that class methods return NotImplementedError
    m = ConcreteModel()
    m.p = ParameterBlock()

    with pytest.raises(NotImplementedError):
        m.p.get_metadata()


@pytest.mark.unit
def test_get_phase_component_set():
    m = ConcreteModel()
    m.p = ParameterBlock()

    m.p.phase_list = Set(initialize=["1", "2", "3"])
    m.p.component_list = Set(initialize=["a", "b", "c"])

    # Need to call this to trigger build of Phase objects for now
    m.p._make_phase_objects()

    pc_set = m.p.get_phase_component_set()

    assert isinstance(m.p._phase_component_set, Set)
    assert len(m.p._phase_component_set) == 9
    for v in m.p._phase_component_set:
        assert v[0] in m.p.phase_list
        assert v[1] in m.p.component_list

    assert pc_set is m.p._phase_component_set

    # Check that method returns existing component
    # Delete phase list so that build will fail to make sure it isn't rebuilding
    m.p.del_component(m.p.phase_list)

    assert m.p.get_phase_component_set() is m.p._phase_component_set


@pytest.mark.unit
def test_get_phase_component_set_subset():
    m = ConcreteModel()
    m.p = ParameterBlock()

    m.p.phase_list = ["1", "2", "3"]
    m.p.phase_comp = {"1": ["a", "b", "c"],
                      "2": ["a", "b"],
                      "3": ["c"]}

    # Need to call this to trigger build of Phase objects for now
    m.p._make_phase_objects()

    pc_set = m.p.get_phase_component_set()

    assert isinstance(m.p._phase_component_set, Set)
    assert len(m.p._phase_component_set) == 6
    for v in m.p._phase_component_set:
        assert v[0] in m.p.phase_comp.keys()
        assert v[1] in m.p.phase_comp[v[0]]

    assert pc_set is m.p._phase_component_set


@pytest.mark.unit
def test_get_component():
    m = ConcreteModel()
    m.p = ParameterBlock()

    m.meta_object = PropertyClassMetadata()

    def get_metadata(self):
        return m.meta_object
    m.p.get_metadata = types.MethodType(get_metadata, m.p)

    with pytest.raises(AttributeError):
        m.p.get_component("foo")

    m.p.comp = Component()

    assert m.p.get_component("comp") is m.p.comp

    m.p.a = object()

    with pytest.raises(
            PropertyPackageError,
            match="p get_component found an attribute a, but it does not "
            "appear to be an instance of a Component object."):
        m.p.get_component("a")


@pytest.mark.unit
def test_get_phase():
    m = ConcreteModel()
    m.p = ParameterBlock()

    with pytest.raises(AttributeError):
        m.p.get_phase("foo")

    m.p.phase = Phase()

    assert m.p.get_phase("phase") is m.p.phase

    m.p.a = object()

    with pytest.raises(
            PropertyPackageError,
            match="p get_phase found an attribute a, but it does not "
            "appear to be an instance of a Phase object."):
        m.p.get_phase("a")


@pytest.mark.unit
def test_make_component_objects():
    m = ConcreteModel()
    m.p = ParameterBlock()

    m.meta_object = PropertyClassMetadata()

    def get_metadata(self):
        return m.meta_object
    m.p.get_metadata = types.MethodType(get_metadata, m.p)

    m.p.component_list = Set(initialize=["comp1", "comp2"])

    m.p._make_component_objects()

    assert isinstance(m.p.comp1, Component)
    assert isinstance(m.p.comp2, Component)


@pytest.mark.unit
def test_make_component_objects_already_exists():
    m = ConcreteModel()
    m.p = ParameterBlock()

    m.p.comp1 = object()

    m.p.component_list = Set(initialize=["comp1", "comp2"])

    with pytest.raises(PropertyPackageError,
                       match="p could not add Component object comp1 - an "
                       "object with that name already exists."):
        m.p._make_component_objects()


@pytest.mark.unit
def test_make_phase_objects():
    m = ConcreteModel()
    m.p = ParameterBlock()

    m.p.phase_list = Set(initialize=["phase1", "phase2"])
    m.p.phase_comp = {"phase1": [1, 2, 3],
                      "phase2": [4, 5, 6]}

    m.p._make_phase_objects()

    assert isinstance(m.p.phase1, Phase)
    assert isinstance(m.p.phase2, Phase)

    assert m.p.phase1.config.component_list == [1, 2, 3]
    assert m.p.phase2.config.component_list == [4, 5, 6]


@pytest.mark.unit
def test_make_phase_objects_already_exists():
    m = ConcreteModel()
    m.p = ParameterBlock()

    m.p.phase1 = object()

    m.p.phase_list = Set(initialize=["phase1", "phase2"])

    with pytest.raises(PropertyPackageError,
                       match="p could not add Phase object phase1 - an "
                       "object with that name already exists."):
        m.p._make_phase_objects()


@pytest.mark.unit
def test_validate_parameter_block_no_component_list():
    m = ConcreteModel()
    m.p = ParameterBlock()

    with pytest.raises(
            PropertyPackageError,
            match="Property package p has not defined a component list."):
        m.p._validate_parameter_block()


@pytest.mark.unit
def test_validate_parameter_block_no_phase_list():
    m = ConcreteModel()
    m.p = ParameterBlock()

    m.meta_object = PropertyClassMetadata()

    def get_metadata(self):
        return m.meta_object
    m.p.get_metadata = types.MethodType(get_metadata, m.p)

    m.p.component_list = Set(initialize=["c1", "c2"])

    with pytest.raises(
            PropertyPackageError,
            match="Property package p has not defined a phase list."):
        m.p._validate_parameter_block()


@pytest.mark.unit
def test_validate_parameter_block_invalid_component_object():
    m = ConcreteModel()
    m.p = ParameterBlock()

    m.p.component_list = Set(initialize=["foo"])

    m.p.phase_list = Set(initialize=["p1", "p2"])
    m.p.foo = object()

    with pytest.raises(TypeError):
        m.p._validate_parameter_block()


@pytest.mark.unit
def test_validate_parameter_block_invalid_phase_object():
    m = ConcreteModel()
    m.p = ParameterBlock()

    m.meta_object = PropertyClassMetadata()

    def get_metadata(self):
        return m.meta_object
    m.p.get_metadata = types.MethodType(get_metadata, m.p)

    m.p.component_list = Set(initialize=["c1", "c2"])

    m.p.phase_list = Set(initialize=["foo"])
    m.p.foo = object()

    with pytest.raises(TypeError):
        m.p._validate_parameter_block()


@pytest.mark.unit
def test_has_inherent_Reactions():
    m = ConcreteModel()
    m.p = ParameterBlock()

    assert not m.p.has_inherent_reactions

    m.p._has_inherent_reactions = True

    assert m.p.has_inherent_reactions


# -----------------------------------------------------------------------------
# Test StateBlock
@declare_process_block_class("TestStateBlock", block_class=StateBlock)
class _StateBlockData(StateBlockData):
    def build(self):
        super(StateBlockData, self).build()


@declare_process_block_class("TestStateBlock2", block_class=StateBlock)
class _StateBlockData2(StateBlockData):
    def build(self):
        super(StateBlockData, self).build()

    def define_state_vars(self):
        return {}


@pytest.mark.unit
def test_StateBlockBase_initialize():
    # Test that StateBlock initialize method raises NotImplementedError
    m = ConcreteModel()
    # Need to index block so that it does not do multiple inheritance
    m.s = Set(initialize=[1, 2])
    m.p = TestStateBlock2(m.s)

    with pytest.raises(NotImplementedError):
        m.p.initialize()


@pytest.mark.unit
def test_StateBlockBase_report():
    # Test that StateBlock initialize method raises NotImplementedError
    m = ConcreteModel()
    # Need to index block so that it does not do multiple inheritance
    m.s = Set(initialize=[0, 1])
    m.p = TestStateBlock2(m.s)

    m.p.report(dof=True)


# -----------------------------------------------------------------------------
# Test StateBlockData
@pytest.mark.unit
def test_StateBlock_config():
    # Test that StateBlockData config has correct arguments
    m = ConcreteModel()
    m.p = TestStateBlock()

    assert len(m.p.config) == 3
    assert hasattr(m.p.config, "has_phase_equilibrium")
    assert hasattr(m.p.config, "defined_state")
    assert hasattr(m.p.config, "parameters")

    m.p.config.has_phase_equilibrium = True
    m.p.config.has_phase_equilibrium = False
    with pytest.raises(ValueError):
        m.p.config.has_phase_equilibrium = 'foo'
    with pytest.raises(ValueError):
        m.p.config.has_phase_equilibrium = 10

    m.p.config.defined_state = True
    m.p.config.defined_state = False
    with pytest.raises(ValueError):
        m.p.config.defined_state = 'foo'
    with pytest.raises(ValueError):
        m.p.config.defined_state = 10


@pytest.mark.unit
def test_StateBlock_NotImplementedErrors():
    # Test that placeholder methods return NotImplementedErrors
    m = ConcreteModel()
    m.p = TestStateBlock()

    with pytest.raises(NotImplementedError):
        m.p.define_state_vars()
    with pytest.raises(NotImplementedError):
        m.p.define_port_members()
    with pytest.raises(NotImplementedError):
        m.p.get_material_flow_terms()
    with pytest.raises(NotImplementedError):
        m.p.get_material_density_terms()
    with pytest.raises(NotImplementedError):
        m.p.get_material_diffusion_terms()
    with pytest.raises(NotImplementedError):
        m.p.get_enthalpy_flow_terms()
    with pytest.raises(NotImplementedError):
        m.p.get_energy_density_terms()
    with pytest.raises(NotImplementedError):
        m.p.get_energy_diffusion_terms()
    with pytest.raises(NotImplementedError):
        m.p.calculate_bubble_point_temperature()
    with pytest.raises(NotImplementedError):
        m.p.calculate_dew_point_temperature()
    with pytest.raises(NotImplementedError):
        m.p.calculate_bubble_point_pressure()
    with pytest.raises(NotImplementedError):
        m.p.calculate_dew_point_pressure()


# -----------------------------------------------------------------------------
# Test parameter block reference attribute
@declare_process_block_class("Parameters")
class _Parameters(PhysicalParameterBlock):
    def build(self):
        super(_Parameters, self).build()

        self.phase_list = ["p1"]
        self.component_list = ["c1"]

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties({'a': {'method': 'a_method'},
                            'recursion1': {'method': '_recursion1'},
                            'recursion2': {'method': '_recursion2'},
                            'not_callable': {'method': 'test_obj'},
                            'raise_exception': {'method': '_raise_exception'},
                            'not_supported': {'method': False},
                            'does_not_create_component': {
                                'method': '_does_not_create_component'}})


@declare_process_block_class("StateTest", block_class=StateBlock)
class _StateTest(StateBlockData):
    def build(self):
        super(_StateTest, self).build()


@pytest.mark.unit
def test_param_ref():
    m = ConcreteModel()
    m.pb = Parameters()
    m.p = StateTest(default={"parameters": m.pb})

    assert m.p.params == m.p.config.parameters


@pytest.mark.unit
def test_validate_params():
    # Test that validate params has been triggered
    m = ConcreteModel()
    m.pb = Parameters()
    m.p = StateTest(default={"parameters": m.pb})

    # If validation has been triggered, Phase & Component objects should exist
    assert isinstance(m.pb.p1, Phase)
    assert isinstance(m.pb.c1, Component)


@pytest.mark.unit
def test_has_inherent_reactions_state_block():
    m = ConcreteModel()
    m.pb = Parameters()
    m.p = StateTest(default={"parameters": m.pb})

    assert not m.p.has_inherent_reactions

    m.pb._has_inherent_reactions = True

    assert m.p.has_inherent_reactions


# -----------------------------------------------------------------------------
# Test properties __getattr__ method
@declare_process_block_class("State", block_class=StateBlock)
class _State(StateBlockData):
    def build(self):
        super(StateBlockData, self).build()

        self.test_obj = 1

    def a_method(self):
        self.a = Var(initialize=1)

    def _recursion1(self):
        self.recursive_cons1 = Constraint(expr=self.recursion2 == 1)

    def _recursion2(self):
        self.recursive_cons2 = Constraint(expr=self.recursion1 == 1)

    def _raise_exception(self):
        raise Exception()

    def _does_not_create_component(self):
        pass


@pytest.fixture()
def m():
    m = ConcreteModel()
    m.pb = Parameters()
    m.p = State(default={"parameters": m.pb})

    return m


@pytest.mark.unit
def test_getattr_add_var(m):
    assert isinstance(m.p.a, Var)
    assert m.p.a.value == 1
    assert m.p.is_property_constructed("a")


@pytest.mark.unit
def test_lock_attributes(m):
    with pytest.raises(AttributeError):
        with m.p.lock_attribute_creation_context():
            m.p.a.value == 1
    # Make sure it unlocked
    assert m.p.a.value == 1


@pytest.mark.unit
def test_is_property_constructed(m):
    assert not m.p.is_property_constructed("a")
    assert m.p.a.value == 1
    assert m.p.is_property_constructed("a")


@pytest.mark.unit
def test_getattr_protected(m):
    with pytest.raises(PropertyPackageError):
        # Call a protected component that does not exist
        m.p.cons = Constraint(expr=m.p._foo == 1)


@pytest.mark.unit
def test_getattr_recursion(m):
    with pytest.raises(PropertyPackageError):
        # Call a component that triggers a recursive loop of calls
        m.p.cons = Constraint(expr=m.p.recursion1 == 1)


@pytest.mark.unit
def test_getattr_does_not_exist(m):
    with pytest.raises(PropertyNotSupportedError):
        m.p.cons = Constraint(expr=m.p.does_not_exist == 1)


@pytest.mark.unit
def test_getattr_not_callable(m):
    with pytest.raises(PropertyPackageError):
        m.p.cons = Constraint(expr=m.p.not_callable == 1)


@pytest.mark.unit
def test_getattr_not_supported(m):
    with pytest.raises(PropertyNotSupportedError):
        m.p.cons = Constraint(expr=m.p.not_supported == 1)


@pytest.mark.unit
def test_getattr_raise_exception(m):
    with pytest.raises(Exception):
        m.p.cons = Constraint(expr=m.p.raise_exception == 1)


# TODO : Need a test for cases where method does not create property
#@pytest.mark.unit
#def test_getattr_does_not_create_component(m):
#    with pytest.raises(PropertyPackageError):
#        m.p.cons = Constraint(expr=m.p.does_not_create_component == 1)



from .fcc_lattice import FCCLattice
from .perovskite_lattice import PerovskiteLattice
from .cubic_lattice import CubicLattice
from .diamond_lattice import DiamondLattice
from .wurtzite_lattice import WurtziteLattice


##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2020, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes-pse".
##############################################################################
import idaes
import logging
import bisect
import threading
from collections.abc import Iterable

from contextlib import contextmanager
from pyomo.common.tee import capture_output


# Throw the standard levels in here, just let you access it all in one place
CRITICAL = logging.CRITICAL # 50
ERROR = logging.ERROR # 40
WARNING = logging.WARNING # 30
INFO_LOW = 21 #Most important info
INFO = logging.INFO # 20  #Medium info (default)
INFO_HIGH = 19 #Less improtant important info
DEBUG = logging.DEBUG # 10
NOTSET = logging.NOTSET # 0


levelname = { # the level name of all our extra info levels is "INFO"
    INFO_HIGH: "INFO",
    INFO_LOW: "INFO",
}


class _TagFilter(logging.Filter):
    """Filter applied to IDAES loggers returned by this modulue."""
    def filter(record):
        """Add in the custom level name and let the record through"""
        if record.levelno in levelname:
            record.levelname = levelname[record.levelno]
        if record.levelno >= WARNING:
            return True
        try:
            if record.tag is None or record.tag in idaes.cfg.logger_tags:
                return True
        except AttributeError:
            return True
        return False


def __info_low(self, *args, **kwargs):
    self.log(INFO_LOW, *args, **kwargs)


def __info_high(self, *args, **kwargs):
    self.log(INFO_HIGH, *args, **kwargs)


def __add_methods(log, tag=None):
    log.addFilter(_TagFilter)
    log = logging.LoggerAdapter(log, {"tag": tag})
    log.info_high = __info_high.__get__(log)
    log.info_low = __info_low.__get__(log)
    # hopefully adding this multiple times is not a problem
    return log


def _getLogger(name, logger_name="idaes", level=None, tag=None):
    assert tag in idaes.cfg.valid_logger_tags.union({None})
    if name.startswith("idaes."):
        name = name[6:]
    name = ".".join([logger_name, name])
    l = logging.getLogger(name)
    if level is not None:
        l.setLevel(level)
    return __add_methods(logging.getLogger(name), tag)


def getIdaesLogger(name, level=None, tag=None):
    """ Return an idaes logger.

    Args:
        name: usually __name__
        level: standard IDAES logging level (default use IDAES config)
        tag: logger tag for filtering, see valid_log_tags()

    Returns:
        logger
    """
    return _getLogger(name=name, logger_name="idaes", level=level, tag=tag)


getLogger = getIdaesLogger

def getSolveLogger(name, level=None, tag=None):
    """ Get a solver logger

    Args:
        name: logger name is "idaes.solve." + name (if name starts with "idaes."
            it is removed before creating the logger name)
        level: Log level
        tag: logger tag for filtering, see valid_log_tags()

    Returns:
        logger
    """
    return _getLogger(name=name, logger_name="idaes.solve", level=level, tag=tag)

def getInitLogger(name, level=None, tag=None):
    """ Get a model initialization logger

    Args:
        name: Object name (usually Pyomo Component name)
        level: Log level
        tag: logger tag for filtering, see valid_log_tags()

    Returns:
        logger
    """
    return _getLogger(name=name, logger_name="idaes.init", level=level, tag=tag)


def getModelLogger(name, level=None, tag=None):
    """ Get a logger for an IDAES model. This function helps users keep their
    loggers in a standard location and use the IDAES logging config.

    Args:
        name: Name (usually __name__).  Any starting 'idaes.' is stripped off, so
            if a model is part of the idaes package, 'idaes' won't be repeated.
        level: Standard Python logging level (default use IDAES config)
        tag: logger tag for filtering, see valid_log_tags()

    Returns:
        logger
    """
    return _getLogger(name=name, logger_name="idaes.model",
                      level=level, tag=tag)


def condition(res):
    """Get the solver termination condition to log.  This isn't a specifc value
    that you can really depend on, just a message to pass on from the solver for
    the user's benefit. Sometimes the solve is in a try-except, so we'll handle
    None and str for those cases, where you don't have a real result."""

    if res is None:
        return "Error, no result"
    elif isinstance(res, str):
        return res
    else:
        s = str(res.solver.termination_condition)

    try:
        if "ipopt" in str(res.solver.message).lower():
            solver_message = " ".join(str(res.solver.message).split(" ")[2:])
            return "{} - {}".format(s, solver_message)
        else:
            return "{} - {}".format(s, str(res.solver.message))
    except:
        return s

def solver_capture_on():
    """This function turns on the solver capture for the solver_log context
    manager. If this is on, solver output within the solver_log context
    is captured and sent to the logger.

    """
    idaes.cfg.logger_capture_solver = True

def solver_capture_off():
    """This function turns off the solver capture for the solver_log context
    manager. If this is off, solver output within the solver_log context
    is just sent to stdout like normal.

    """
    idaes.cfg.logger_capture_solver = False

def solver_capture():
    """Return True if solver capture is on or False otherwise."""
    return idaes.cfg.logger_capture_solver

def log_tags():
    """Returns a set of logging tags to be logged.

    Returns:
        (set) tags to be logged
    """
    return idaes.cfg.logger_tags

def set_log_tags(tags):
    """Specify a set of tags to be logged

    Args:
        tags(iterable of str): Tags to log

    Returns:
        None
    """
    for m in tags:
        if m not in idaes.cfg.valid_logger_tags.union({None}):
            raise ValueError("{} is not a valid logging tag".format(m))
    idaes.cfg.logger_tags = set(tags)

def add_log_tag(tag):
    """Add a tag to the list of tags to log.

    Args:
        tag(str): Tag to log

    Returns:
        None
    """
    if tag not in idaes.cfg.valid_logger_tags.union({None}):
        raise ValueError("{} is not a valid logging tag".format(tag))
    idaes.cfg.logger_tags.add(tag)

def remove_log_tag(tag):
    """Remove a tag from the list of tags to log.

    Args:
        tag(str): Tag to no longer log

    Returns:
        None
    """
    try:
        idaes.cfg.logger_tags.remove(tag)
    except ValueError:
        pass

def valid_log_tags():
    """Returns a set of valid logging tag names.

    Returns:
        (set) valid tag names
    """
    return idaes.cfg.valid_logger_tags.union({None})

def add_valid_log_tag(tag):
    """Add a tag name to the list of valid names.

    Args:
        tag(str): A tag name

    Returns:
        None
    """
    assert isinstance(tag, str)
    idaes.cfg.valid_logger_tags.add(tag)


class IOToLogTread(threading.Thread):
    """This is a Thread class that can log solver messages and show them as
    they are produced, while the main thread is waiting on the solver to finish
    """

    def __init__(self, stream, logger, sleep=1.0, level=logging.ERROR):
        super().__init__(daemon=True)
        self.log = logger
        self.level = level
        self.stream = stream
        self.sleep = sleep
        self.stop = threading.Event()
        self.pos=0

    def log_value(self):
        try:
            v = self.stream.getvalue()[self.pos:]
        except ValueError:
            self.stop.set()
            return
        self.pos += len(v)
        for l in v.split("\n"):
            if l:
                self.log.log(self.level, l.strip())

    def run(self):
        while True:
            self.log_value()
            self.stop.wait(self.sleep)
            if self.stop.isSet():
                self.log_value()
                self.pos=0
                return


class SolverLogInfo(object):
    def __init__(self, tee=True, thread=None):
        self.tee = tee
        self.thread = thread


@contextmanager
def solver_log(logger, level=logging.ERROR):
    """Context manager to send solver output to a logger. This uses a separate
    thread to log solver output while the solver is running"""
    # wait 3 seconds to  join thread.  Should be plenty of time.  In case
    # something goes horribly wrong though don't want to hang.  The logging
    # thread is daemonic, so it will shut down with the main process even if it
    # stays around for some mysterious reason while the model is running.
    join_timeout = 3
    tee = logger.isEnabledFor(level)
    if not solver_capture():
        yield SolverLogInfo(tee=tee)
    else:
        with capture_output() as s:
            lt = IOToLogTread(s, logger=logger, level=level)
            lt.start()
            try:
                yield SolverLogInfo(tee=tee, thread=lt)
            except:
                lt.stop.set()
                lt.join(timeout=join_timeout)
                raise
        # thread should end when s is closed, but setting stop makes sure
        # the last of the output gets logged before closing s
        lt.stop.set()
        lt.join(timeout=join_timeout)
